# kristine Larson, February 2022
# library to support RINEX downloads.
# started with RINEX 3, and now adding RINEX 2
# this should replace the spaghetti code in gps.py
import datetime
import json
import numpy as np
import os
import requests
import sys
import subprocess
import time
import wget
from urllib.parse import urlparse
# local libraries
import gnssrefl.gps as g
import gnssrefl.highrate as ch
import gnssrefl.kelly as kelly

def gogetit(dir1, filename, ext):
    """
    The purpose of this function is to to download RINEX 2 files.
    The code will try to get the file and check to see if it was successful.

    Parameters
    ----------
    dir1 : str
        the main https directory address 
    filename : str
        name of the GNSS files
    ext : str
        kind of ending to the filename, (Z, gz etc)

    Returns 
    -------
    foundit : bool
        whether file was found
    f : str
        name of the file
    """
    foundit = False
    f= filename + ext
    print(dir1 + f)
    if os.path.exists(f):
        print('File is already on your disk')
        foundit = True
    else:
        try:
            wget.download(dir1+f,f)
        except:
            okok =1
        if os.path.exists(f):
            print('\n File has been found ',f)
            foundit = True

    return foundit, f

def swapRS(stream):
    """
    function that swaps R to S and vice versa for RINEX 3 files

    Parameters
    ----------
    stream : str
        RINEX 3 filename streaming acronym (S or R)

    Returns
    -------
    newstream : str
        the opposite of what was in stream

    """
    if stream == 'R':
        newstream = 'S'
    else:
        newstream = 'R'
    return newstream

def just_bkg(cyyyy, cdoy, file_name):
    """

    looks for RINEX files at BKG in two directories

    Parameters
    ----------
    cyyyy:  str
        four character year

    cdoy: str
        three character day of year

    file_name: str
        expected RINEX filename

    """
    # BKG has two separate archives - 
    dir1 = 'https://igs.bkg.bund.de/root_ftp/IGS/obs/' + cyyyy + '/' + cdoy + '/'
    dir2 = 'https://igs.bkg.bund.de/root_ftp/EUREF/obs/' + cyyyy + '/' + cdoy + '/'
    print('>>> Trying first : ',dir1+file_name)
    fexist = False
    try:
        fexist = g.replace_wget(dir1+file_name, file_name)
        #wget.download(dir1+file_name,file_name)
        #fexist = True
    except:
        print('did not find the RINEX 3 file')
    if (not fexist):
        print('>>> Now trying: ',dir2+file_name)
        try:
            fexist = g.replace_wget(dir2+file_name, file_name)
            #wget.download(dir2+file_name,file_name)
            #fexist = True
        except:
            print('did not find it there either')

    return

def universal(station9ch, year, doy, archive,srate,stream,debug=False):
    """
    main code for seamless archive for RINEX 3 files ... 

    Parameters
    ----------
    station9ch : str
        nine character station name

    year : int
        year 

    doy : int
        day of year

    archive : str
        archive name

    srate : int
        receiver samplerate

    stream : str
        one character: R or S

    debug : bool
        whether debugging statements printed

    Returns
    -------
    file_name : str
        name of rinexfile

    foundit : boolean
        whether file was found

    """
    # define the file name
    if debug:
        print('Searching the ', archive, ' archive with rate/filetype', srate, stream)
    foundit = False
    dtmp, month, day, cyyyy,cdoy, YMD = g.ydoy2useful(year,doy)
    cmm = '{:02d}'.format(month)
    cdd = '{:02d}'.format(day)

    file_name,cyyyy,cdoy = filename_plus(station9ch,year,doy,srate,stream)
    print('Filename:', file_name)
    cyy = cyyyy[2:4]
    dir1 = ''
    dir2 = ''

    # put this outside the try because I think there is one in the function
    if archive == 'bfg':
        station = station9ch[0:4] ; samplerate = srate; stream = 'R'
        debug = False
        g.bfg_data(station, year, doy, samplerate,debug)
        zip_file_name = file_name[0:-2] + 'zip'
        if os.path.exists(file_name):
            if debug:
                print('File was found: ', file_name )
            foundit = True
        elif os.path.exists(zip_file_name):
            if debug:
                print('From zip to gz: ', zip_file_name )
            subprocess.call(['unzip', zip_file_name])
            subprocess.call(['gzip', file_name[0:-3]])
            # cleaning up
            subprocess.call(['rm', zip_file_name])
            foundit = True
        else:
            print('File was not found: at bfg.')

        return file_name,foundit
    cydoy =  cyyyy + '/' + cdoy + '/'
    s1 = time.time()
    if (archive == 'unavco'):
        # original dir1 = 'https://data.unavco.org/archive/gnss/rinex3/obs/' + cyyyy + '/' + cdoy + '/'
        #url1 =      'https://data-idm.unavco.org/archive/gnss/rinex3/obs/' + cydoy + file_name
        url1 =      'https://data.unavco.org/archive/gnss/rinex3/obs/' + cydoy + file_name
        foundit,file_name = kelly.the_kelly_simple_way(url1,file_name)
        s2 = time.time()
        if debug:
            print('Download took ',np.round(s2-s1,2), ' seconds') 
        return file_name,foundit

    try:
        if (archive == 'ign'):
            dir1='ftp://igs.ensg.ign.fr/pub/igs/data/' + cyyyy + '/' + cdoy + '/'
            wget.download(dir1+file_name,file_name)
        elif (archive == 'sonel'):
            dir1='ftp://ftp.sonel.org/gps/data/' + cyyyy + '/' + cdoy + '/'
            wget.download(dir1+file_name,file_name)
        elif archive == 'bkg':
            just_bkg(cyyyy, cdoy, file_name)
        elif (archive == 'bev'):
            dir1 = 'https://gnss.bev.gv.at/at.gv.bev.dc/data/obs/' + cyyyy + '/' + cdoy + '/'
            #wget.download(dir1+file_name,file_name)
            fex = g.replace_wget(dir1+file_name, file_name)
        elif (archive == 'epn'):
            dir1 = 'https://epncb.oma.be/ftp/obs/' + cyyyy + '/' + cdoy + '/'
            #print(dir1)
            wget.download(dir1+file_name,file_name)
        elif (archive == 'ignes'):
           # if someone wanted to help by adding this to the high rate
           # https://datos-geodesia.ign.es/ERGNSS/horario_1s/YYYYMMDD/HH/file
            dir1 = 'https://datos-geodesia.ign.es/ERGNSS/diario_30s/' + cyyyy + '/' + cyyyy + cmm + cdd + '/'
            #print(dir1)
            wget.download(dir1+file_name,file_name)
        elif (archive == 'ga'):
            QUERY_PARAMS, headers = ga_stuff(station9ch, year, doy)
            API_URL = 'https://data.gnss.ga.gov.au/api/rinexFiles/'  
            request = requests.get(API_URL, QUERY_PARAMS, headers=headers)
        #request.raise_for_status()
        # i don't know how to read this - so doing it the dumb way
            if len(json.loads(request.content)) == 1:
                for query_response_item in json.loads(request.content):
                    file_url = query_response_item['fileLocation']
                    file_name = urlparse(file_url).path.rsplit('/', 1)[1]
                    fex = g.replace_wget(file_url, file_name)
                    #wget.download(file_url,file_name)
        elif (archive == 'nrcan'):
            dir1 = 'https://cacsa.nrcan.gc.ca/gps/data/gpsdata/' + cyy + cdoy  + '/' + cyy + 'd' + '/'
            wget.download(dir1+file_name,file_name)
        #elif (archive == 'unavco-old'):
        #    dir1 = 'https://data.unavco.org/archive/gnss/rinex3/obs/' + cyyyy + '/' + cdoy + '/'
        #    wget.download(dir1+file_name,file_name)
        elif (archive == 'gfz'):
            dir1 = 'ftp://isdcftp.gfz-potsdam.de/gnss/data/daily/' + cyyyy + '/' + cdoy + '/'
            wget.download(dir1+file_name,file_name)
        elif (archive == 'cddis'):
            new_way_dir = '/gnss/data/daily/' + cyyyy + '/' + cdoy + '/' + cyy + 'd/'
            g.cddis_download_2022B(file_name,new_way_dir)
        else:
            return '', ''
    except:
        okokok = 1
    s2 = time.time()
    if debug: 
        print('Download took ',np.round(s2-s1,2), ' seconds') 
    if os.path.exists(file_name):
        siz = os.path.getsize(file_name)
        if (siz == 0):
            subprocess.call(['rm',file_name])
            print('File was not found: ', file_name, ' at ', archive)
        else:
            print('File exists: ', file_name, ' at ', archive)
            foundit = True
    else:
        print('File was not found: ', file_name, ' at ', archive)

    return file_name,foundit

def filename_plus(station9ch,year,doy,srate,stream):
    """
    function to create RINEX 3 filenames for one day files.

    Parameters
    ----------
    station9ch : str
        9 character station name
    year : int
        full year
    doy : int
        day of year
    srate : int
        receiver sample rate 
    stream : str 
        R or S ; latter means the file was streamed.
    output : str
        compliant filename with crx.gz on the end as this is how the files 
        are stored at GNSS archives as far as I know.

    Returns
    -------
    file_name : str
        filename of the RINEX 3 file
    cyyyy : str
        4ch year 
    cdoy : str
        3ch day of year

    """
    cyyyy = str(year)
    cdoy = '{:03d}'.format(doy)
    crate = '{:02d}'.format(srate)
    upper = station9ch.upper()
    ff = upper + '_' + stream + '_' + cyyyy + cdoy + '0000_01D_' + crate + 'S_MO'
    ext = '.crx.gz'
    file_name = ff + ext

    return file_name, cyyyy, cdoy

def ga_stuff(station, year, doy,rinexv=3):
    """
    GA API requirements to download a Rinex 3 file

    Parameters
    ----------
    station : str
        9 character station name
    year : integer
        full year
    doy : int
        day of year
    rinexv : int
        rinex version        

    Returns
    -------
    QUERY_PARAMS : dict
        I think
    headers : dict
        I think

    """
    d = datetime.datetime(year, 1, 1) + datetime.timedelta(days=(doy-1))
    month = int(d.month); day = int(d.day)
    cmonth = '{:02d}'.format(month); cday = '{:02d}'.format(day)


    QUERY_PARAMS = {}
    QUERY_PARAMS['stationId'] = station[0:4].upper()
    QUERY_PARAMS['fileType'] = 'obs'
    QUERY_PARAMS['filePeriod'] = '01D'
    QUERY_PARAMS['rinexVersion'] = str(rinexv) 
    QUERY_PARAMS['startDate'] = str(year) + '-' + cmonth + '-' + cday + 'T00:00:00Z'
    QUERY_PARAMS['endDate'] =   str(year) + '-' + cmonth + '-' + cday + 'T23:59:30Z'

    headers = {}
    headers['Accept-Encoding'] = 'gzip'

    return QUERY_PARAMS, headers

def universal_all(station9ch, year, doy, srate,stream,screenstats):
    """
    check multiple archives for RINEX 3 data

    Parameters
    ----------
    station9ch : str
        9 character station name
    year : int
        full year
    doy : int
        doy of year
    srate : int
        receiver sample rate 
    stream : str
        R or S
    screenstats : bool
        print statements

    Peturns
    -------
    file_name : str
        rinex filename
    foundit : bool
        whether rinex file was found

    """
    foundit = False

    for archive in ['unavco','cddis','bkg','bev','epn','ga']:
        if archive == 'unavco':
            srate_in = 15
        else:
            srate_in = srate
        # only look if you haven't found it
        if (not foundit):
            file_name,dd = universal(station9ch, year, doy, archive,srate_in,stream)
            if os.path.exists(file_name):
                foundit = True

    return file_name, foundit

def rinex2names(station,year,doy):
    """
    Creates the expected filename for rinex2 version files

    Parameters
    ----------
    station : string

    year : integer

    doy : integer
        day of year

    Results
    -------
    f1 : str
        hatanaka rinex filename
    f2 : str
        regular rinex filename
    cyyyy : str
        four character year 

    cdoy : string
        three character day of year

    """
    cyyyy = str(year)
    cdoy = '{:03d}'.format(doy)
    f1 = station  + cdoy + '0.' + cyyyy[2:4] + 'd'
    f2 = station  + cdoy + '0.' + cyyyy[2:4] + 'o'

    return f1, f2, cyyyy, cdoy


def universal_rinex2(station, year, doy, archive,screenstats):
    """
    The long-awaited seamless archive for rinex 2 files ...

    Parameters
    ----------
    station : str
        four character station name
    year : int
        full year
    doy : int
        day of year
    archive : str
        name of the GNSS archive
    screenstats : bool
        whether print statements come to scree

    Returns
    -------
    file_name : str
        RINEX filename that was downloaded
    foundit : bool
        whether file was found

    """
    # define the file name

    foundit = False; dir1 = ''; file_name = ''

    dname, oname, cyyyy, cdoy = rinex2names(station,year,doy)
    if os.path.exists(oname):
        #print('RINEX o File is already on disk')
        return oname, True 

    if screenstats:
        print('Searching the ', archive, ' archive for ', station)

    cydoy = cyyyy + '/' + cdoy + '/'
    cyy = cyyyy[2:4]

    s1 = time.time()
    if (archive == 'jp'):
        # i did not want to rewrite the code
        gsi_data(station, year, doy)
        file_name = oname
        if os.path.exists(file_name):
            foundit = True
    #elif (archive == 'unavco-old'):
    #    dir1 = 'https://data.unavco.org/archive/gnss/rinex/obs/' + cydoy
    #    foundit, file_name = gogetit(dir1, dname, '.Z'); 
    #    if not foundit:
    #        foundit, file_name = gogetit(dir1, oname, '.Z')
    elif (archive == 'unavco'):
        #url1 = 'https://data-idm.unavco.org/archive/gnss/rinex/obs/' + cydoy + dname + '.Z'
        url1 = 'https://data.unavco.org/archive/gnss/rinex/obs/' + cydoy + dname + '.Z'
        if screenstats:
            print(url1)
        foundit,file_name = kelly.the_kelly_simple_way(url1, dname + '.Z')
        if not foundit:
            url2 = 'https://data.unavco.org/archive/gnss/rinex/obs/' + cydoy + oname + '.Z'
            if screenstats:
                print(url2)
            foundit,file_name = kelly.the_kelly_simple_way(url2, oname + '.Z')
    elif (archive == 'special'):
        #print('testing out new protocol at unavco')
        url1 = 'https://data.unavco.org/archive/gnss/products/reflectometry/' + cydoy + oname + '.gz'
        if screenstats:
            print(url1)
        foundit,file_name = kelly.the_kelly_simple_way(url1, oname + '.gz')
        # old way
        #foundit, file_name = gogetit(dir1, oname, '.gz'); 
    elif archive == 'sopac':
        dir1 = 'ftp://garner.ucsd.edu/pub/rinex/' + cydoy
        file_name = dname + '.Z'
        url = dir1 + '/' + file_name
        foundit, file_name = gogetit(dir1, dname, '.Z');
    elif archive == 'sonel':
        dir1 = 'ftp://ftp.sonel.org/gps/data/' + cydoy
        foundit, file_name = gogetit(dir1, dname, '.Z');
    elif archive == 'nz':
        dir1 =  'https://data.geonet.org.nz/gnss/rinex/' + cydoy + '/'
        # try using requests
        file_name = oname + '.gz'
        url = dir1 + file_name
        foundit = g.replace_wget(url, file_name)
        #foundit, file_name = gogetit(dir1, oname, '.gz'); 
    elif archive == 'bkg':
        # are the old data with Z instead of gz?
        dir1 = 'https://igs.bkg.bund.de/root_ftp/IGS/obs/' + cydoy
        dir2 = 'https://igs.bkg.bund.de/root_ftp/EUREF/obs/' + cydoy

        foundit, file_name = gogetit(dir1, dname, '.Z');
        if not os.path.exists(file_name):
            foundit, file_name = gogetit(dir2, dname, '.Z')
        # not even sure this is a thing ???
        if not os.path.exists(file_name):
            foundit, file_name = gogetit(dir1, dname, '.gz')
        if not os.path.exists(file_name):
            foundit, file_name = gogetit(dir2, dname, '.gz')
    elif (archive == 'bev'):
        dir1 = 'https://gnss.bev.gv.at/at.gv.bev.dc/data/obs/' + cydoy
        foundit, file_name = gogetit(dir1, dname, '.gz');
    elif (archive == 'jeff'):
        if station == 'pbay':
            dir1 = 'ftp://gps.alaska.edu/pub/gpsdata/permanent/C2/' + cydoy
            foundit, file_name = gogetit(dir1, oname, '.gz');
        else:
            dir1 = 'ftp://gps.alaska.edu/pub/gpsdata/CoopCORS/' + cydoy + '/' + station.upper() + '/'
            foundit, file_name = gogetit(dir1, oname, '.gz');
    elif (archive == 'ngs_hourly'):
        print('Hourly NGS')
        delete_hourly= True
        yy,mm,dd = g.ydoy2ymd(year,doy)
        file_name, foundit = g.big_Disk_work_hard(station,yy,mm,dd,delete_hourly)
    elif (archive == 'ngs'):
        dir1 = 'https://geodesy.noaa.gov/corsdata/rinex/' + cydoy + '/' + station + '/'
        file_name = oname + '.gz' ; url = dir1 + file_name
        foundit = g.replace_wget(url,file_name)
        #foundit, file_name = gogetit(dir1, oname, '.gz');
        if not foundit:
            file_name = dname + '.gz'; url = dir1 + file_name
            foundit = g.replace_wget(url,file_name)
            #foundit, file_name = gogetit(dir1, dname, '.gz')
    elif (archive == 'cddis'):
        foundit,file_name = serial_cddis_files(dname,cyyyy,cdoy); 
    elif (archive == 'ga'):
        QUERY_PARAMS, headers = ga_stuff(station, year, doy)
        API_URL = 'https://data.gnss.ga.gov.au/api/rinexFiles/'
        QUERY_PARAMS['rinexVersion'] = '2'
        request = requests.get(API_URL, QUERY_PARAMS, headers=headers)
        file_name = ''
        if len(json.loads(request.content)) == 1:
            foundit = True
            for query_response_item in json.loads(request.content):
                file_url = query_response_item['fileLocation']
                file_name = urlparse(file_url).path.rsplit('/', 1)[1]
                wget.download(file_url,file_name)
    elif (archive == 'nrcan'):
        dir1 = 'ftp://cacsa.nrcan.gc.ca/gps/data/gpsdata/' + cyy + cdoy  + '/' + cyy + 'd' + '/'
        foundit, f = gogetit(dir1, dname, '.Z'); file_name = f
    else:
         print('I do not recognize your archive', archive)

    s2 = time.time()
    if screenstats:
        print('Download took ', np.round(s2-s1,2),' seconds') 

    if not os.path.exists(file_name) and screenstats:
        print('Did not find the Rinex file')

    return file_name,foundit

def make_rinex2_ofiles(file_name):
    """
    take a rinex2 file, gunzip or uncompress it, and 
    then Hatanaka decompress it

    Parameters
    ----------
    file_name: string
        rinex2 filename

    Returns
    -------
    new_name : string
        filename after multiple decompression processes
    fexist : boolean
        whether file was successfully created

    """
    if (file_name[-1:] == 'Z'):
        subprocess.call(['uncompress', file_name])
        file_name = file_name[:-2]
    if (file_name[-2:] == 'gz'):
        subprocess.call(['gunzip', file_name])
        file_name = file_name[:-3]

    crnxpath = g.hatanaka_version()
    # take off the d and make it an o
    new_name = file_name[:-1] + 'o'

    if file_name[-1:] == 'd':
        if os.path.exists(crnxpath):
            subprocess.call([crnxpath, file_name])
            subprocess.call(['rm', '-f',file_name])
        else:
            g.hatanaka_warning()
    elif file_name[-1:] == 'o':
        thisisagoodthing = 1
    else:
        print('I only recognize d and o files')

    fexist = False
    if os.path.exists(new_name):
        fexist = True

    return new_name, fexist

def strip_rinexfile(rinexfile):
    """
    uses either teqc or gfzrnx to reduce observables,
    i.e. only SNR data.

    Parameters
    ----------
    rinexfile : string
        name of the rinex2 file
    """
    print('You chose the strip option to reduce the number of observables in the RINEX file')
    teqcv = g.teqc_version()
    if os.path.isfile(teqcv):
        #print('Use teqc to strip the RINEX 2 file')
        foutname = 'tmp.' + rinexfile
        fout = open(foutname,'w')
        subprocess.call([teqcv, '-O.obs','S1+S2+S5+S6+S7+S8', rinexfile],stdout=fout)
        fout.close()
        subprocess.call(['rm','-f',rinexfile])
        subprocess.call(['mv','-f',foutname, rinexfile])
    else:
        gfzrnxpath = g.gfz_version()
        if os.path.isfile(gfzrnxpath):
            #print('Use gfzrnx to strip the RINEX 2 file')
            tempfile = rinexfile + '.tmp'
            # save yourself heartache down the way cause those doppler data are just clogging up the works
            subprocess.call([gfzrnxpath,'-finp', rinexfile, '-fout', tempfile, '-vo','2','-f', '-obs_types', 'S','-q'])
            subprocess.call(['mv', '-f', tempfile, rinexfile])

def gsi_data(station,year,doy):
    """
    get data from GSI

    Parameters
    ----------
    station : str
        6 char station name
    year : int
        full year
    doy : int
        day of yare

    """
    d = g.doy2ymd(year,doy);
    month = d.month; day = d.day
    g.rinex_jp(station, year, month, day)


def rinex2_highrate(station, year, doy,archive,strip_snr):
    """
    kluge to download highrate data since i have revamped the rinex2 code
    strip_snr is boolean as to whether you want to strip out the non-SNR data
    it can be slow with highrate data. it requires gfzrnx

    Parameters
    ----------
    station : string
         4 character station ID.  lowercase

    year : integer
        full year

    doy : integer
        day of year

    archive : string
        name of GNSS archive

    strip_snr : boolean
        whether you want to strip out the observables (leaving only SNR)

    """
    foundit = False
    d = g.doy2ymd(year,doy);
    month = d.month; day = d.day
    rinexfile,rinexfiled = g.rinex_name(station, year, month, day)
    if (archive == 'unavco') or (archive == 'all'):
        g.rinex_unavco_highrate(station, year, month, day)
    # file does not exist, so keep looking
    if not os.path.isfile(rinexfile):
        if (archive == 'nrcan') or (archive == 'all'):
            g.rinex_nrcan_highrate(station, year, month, day)
    # try new cddis for highrate rINex 2
    if not os.path.isfile(rinexfile):
        if (archive == 'cddis') or (archive == 'all'):
            stream = 'R'
            srate = 1 # one second
            ch.cddis_highrate(station, year, month, day,stream,srate)
    #if not os.path.isfile(rinexfile):
    #    if not os.path.isfile(rinexfile):
    #        if (archive == 'ga') or (archive == 'all'):
    #            g.rinex_ga_highrate(station, year, month, day)

    if os.path.isfile(rinexfile):
        foundit = True
        gfzrnxpath = g.gfz_version()
        if os.path.isfile(gfzrnxpath) and strip_snr:
            print('You chose the strip_snr option to reduce the number of observables')
            tempfile = rinexfile + '.tmp'
        # save yourself heartache down the way cause those doppler data are just clogging up the works
            subprocess.call([gfzrnxpath,'-finp', rinexfile, '-fout', tempfile, '-vo','2','-f', '-obs_types', 'S','-q'])
            subprocess.call(['mv', '-f', tempfile, rinexfile])

    return rinexfile, foundit

def ga_stuff_highrate(station, year, doy,rinexv=3):
    """
    This should be merged with existing ga_stuff code

    Parameters
    ----------
    station : str
        4 or 9 ch station name
    year : int
        full year
    doy : int
        day of year
    rinexv : int
        rinex version

    Returns
    -------
    QUERY_PARAMS : json
        i think
    headers : json
        i think
    """
    d = datetime.datetime(year, 1, 1) + datetime.timedelta(days=(doy-1))
    month = int(d.month); day = int(d.day)
    cmonth = '{:02d}'.format(month); cday = '{:02d}'.format(day)


    QUERY_PARAMS = {}
    QUERY_PARAMS['stationId'] = station[0:4].upper()
    QUERY_PARAMS['fileType'] = 'obs'
    QUERY_PARAMS['filePeriod'] = '15M'

    QUERY_PARAMS['rinexVersion'] = str(rinexv) 
    QUERY_PARAMS['startDate'] = str(year) + '-' + cmonth + '-' + cday + 'T00:00:00Z'
    QUERY_PARAMS['endDate'] =   str(year) + '-' + cmonth + '-' + cday + 'T23:59:30Z'

    headers = {}
    headers['Accept-Encoding'] = 'gzip'

    return QUERY_PARAMS, headers

def serial_cddis_files(dname,cyyyy,cdoy):
    """
    Looks for rinex files in the hatanaka decompression section of cddis

    Parameters
    ----------
    dname : string
        rinex2 filename without compression extension
    cyyyy : string
        four character year
    cdoy : string
        three character day of yaer

    Returns
    -------
    foundit : bool
        whether file was found
    f : str
        filename

    """
    foundit = False
    f = ''
    # try gz, then Z.  so annoying
    gz_file_name = dname + '.gz'
    Z_file_name = dname + '.Z'
    new_way_dir = '/gnss/data/daily/' + cyyyy + '/' + cdoy + '/' + cyyyy[2:4] + 'd/'
    print(gz_file_name, new_way_dir)
    try:
        g.cddis_download_2022B(gz_file_name,new_way_dir) ;
    except:
        okok = 1

    if os.path.exists(gz_file_name):
        siz = os.path.getsize(gz_file_name)
        if siz == 0:
            subprocess.call(['rm',gz_file_name])
        else:
            foundit = True
            f = gz_file_name

    if (not foundit):
        print(Z_file_name, new_way_dir)
        # look for Z compressed file
        try:
            g.cddis_download_2022B(Z_file_name,new_way_dir) ;
        except:
            okok = 1

        if os.path.exists(Z_file_name):
            siz = os.path.getsize(Z_file_name)
            if siz == 0:
                subprocess.call(['rm',Z_file_name])
            else:
                foundit = True
                f = Z_file_name

    return foundit, f



# -*- coding: utf-8 -*-
"""
command line tool for the rinex2snr module
it translates rinex files and makes SNR files

compile the fortran first
f2py -c -m gnssrefl.gpssnr gnssrefl/gpssnr.f

"""

import argparse
import os
import subprocess
import sys
import time

import gnssrefl.gps as g
import gnssrefl.rinex2snr as rnx
import gnssrefl.gnssir_v2 as guts2

from gnssrefl.utils import validate_input_datatypes, str2bool

import multiprocessing
from functools import partial


def parse_arguments():

    #msg = rnx.print_archives()
    g.print_version_to_screen()

    parser = argparse.ArgumentParser()
    #parser = argparse.ArgumentParser(epilog=msg)
    parser.add_argument("station", help="station name", type=str)
    parser.add_argument("year", help="year", type=int)
    parser.add_argument("doy", help="start day of year", type=int)
    parser.add_argument("-snr", default=None, help="snr file ending, 99: 5-30 deg.; 66: < 30 deg.; 88: all data; 50: < 10 deg.", type=int)
    parser.add_argument("-orb", default=None, type=str,
                        help="orbit type, e.g. gps, gps+glo, gnss, rapid, ultra, gnss3")
    parser.add_argument("-rate", default=None, metavar='low', type=str, help="low or high (tells code which folder to search).  If samplerate is 1, this is set automatically to high.") 
    parser.add_argument("-dec", default=None, type=int, help="decimate (seconds)")
    parser.add_argument("-nolook", default=None, metavar='False', type=str,
                        help="True means only use RINEX files on local machine")
    # remove fortran as an option
    #parser.add_argument("-fortran", default=None, metavar='False', type=str,
    #                    help="True means use Fortran RINEX translators ")
    parser.add_argument("-archive", default=None, metavar='all',
                        help="specify archive", type=str)
    parser.add_argument("-doy_end", default=None, help="end day of year", type=int)
    parser.add_argument("-year_end", default=None, help="end year", type=int)
    parser.add_argument("-overwrite", default=None, help="boolean", type=str)
    parser.add_argument("-translator", default=None, help="translator(fortran,hybrid,python)", type=str)
    parser.add_argument("-samplerate", default=None, help="sample rate in sec (RINEX 3 only)", type=int)
    parser.add_argument("-stream", default=None, help="Set to R or S (RINEX 3 only)", type=str)
    parser.add_argument("-mk", default=None, help="use T for uppercase station names and non-standaard archives", type=str)
    parser.add_argument("-weekly", default=None, help="use T for weekly data translation", type=str)
    parser.add_argument("-monthly", default=None, help="use T for monthly data translation", type=str)
    parser.add_argument("-strip", default=None, help="use T to reduce number of obs", type=str)
    parser.add_argument("-screenstats", default=None, help="set to T see more info printed to screen", type=str)
    parser.add_argument("-gzip", default=None, help="boolean, default is SNR files are gzipped after creation", type=str)
    parser.add_argument("-par", default=None, help="parallel processes allowed", type=int)
    parser.add_argument("-timeout", default=None, help="timeout in secs, useful for some archives", type=int)

    args = parser.parse_args().__dict__

    # convert all expected boolean inputs from strings to booleans
    boolean_args = ['nolook', 'overwrite', 'mk', 'weekly','strip','screenstats','gzip','monthly']
    args = str2bool(args, boolean_args)

    # only return a dictionary of arguments that were added from the user - all other defaults will be set in code below
    return {key: value for key, value in args.items() if value is not None}


def rinex2snr(station: str, year: int, doy: int, snr: int = 66, orb: str = None, rate: str = 'low', dec: int = 0,
              nolook: bool = False, archive: str = 'all', doy_end: int = None,
              year_end: int = None, overwrite: bool = False, translator: str = 'hybrid', samplerate: int = 30,
              stream: str = 'R', mk: bool = False, weekly: bool = False, strip: bool = False, 
              screenstats : bool = False, gzip : bool = True, monthly : bool = False, 
              par : int=None, timeout : int = 0 ):
    """
    rinex2snr translates RINEX files to a new file in SNR format. This function will also fetch orbit files for you.
    RINEX obs files are provided by the user or fetched from a long list of archives. Although RINEX 3 is supported, 
    the default is RINEX 2.11 files

    beta version of parallel processing available in this release.  Set -par to a number < 11 
    Some archives have been set to non-compliant with this feature. Please look in the first few lines
    of code to see the names of these archives.

    In general, you should not make RINEX 2.11 files with a huge number of observables. Especially do not put
    Doppler data in your file.  If you have more than 25 observables (multi-GNSS) or 20 (GPS only), the code
    should send an error message to a log. The location of that log is printed to the screen. If you don't want
    to remake your RINEX files, you can try the -strip T option, which I believe uses gfzrnx to strip out everything
    except for SNR data.

    Real-time users should use ultra, wum, or wum2

    Default orbits are GPS only until day of year 137, 2021 when rapid GFZ orbits became available.  If you still want to use
    the nav message, i.e. GPS only, you can request it.

    bkg no longer a boolean input - must be specified with archive name, i.e. bkg-igs or bkg-euref

    For the nolook option :

    If you have the RINEX 2.11 file, the file was originally required to be normal RINEX (ends in o) or 
    gzipped normal RINEX. It can be in the local directory which is where you are running the code 
    or it can be in $REFL_CODE/YYYY/rinex/ssss, where ssss is the lowercase directory name for your station. 
    nolook now allows RINEX 2.11 files that are Hatanaka compressed, Hatanaka compressed + unix compressed,  for 
    the local directory. It also allows Hatanaka compressed in the REFL_CODE directory.  

    If you are running the Docker, it can be a bit confusing to figure out where to put the files.  Please 
    see the discussion in the Docker installation section, as this is my best effort to help you with this.

    Beyond that, you can try the -mk T option which searches other places, i.e. $REFL_CODE/rinex/ etc. I do not recommend
    that you use this option, but it is there.

    For RINEX 3 files, I believe it checks for crx.gz, rnx, or rnx.gz endings in the local directory. It 
    also checks the $REFL_CODE/YYYY/rinex directory for the crx.gz and rnx versions. 
    It looks like I do not delete the RINEX 3 files (though I do delete the RINEX 2.11 files).

    FAQ: what is rate anad srate?  rate is telling the code which folder to use because archives always have 
    files in different directories depending on sample rate.  srate is for RINEX 3 files only because RINEX 3 
    has the sample rate on the filename itself (not just the directory).  

    What is the stream parameter? It is a naming convention that is only used by RINEX 3 people. The allowed 
    file types are S or R.  I believe S stands for streamed.

    RINEX3 30 second archives supported  
        bev, bkg-euref, bkg-igs, cddis, epn, ga, gfz, nrcan, sonel

    RINEX3 15 sec archives
        bfg, unavco  - you may need to specify 15 second sample rate

    RINEX3 1 sec 
        bkg-igs, bkg-euref, cddis, ignes (spain), maybe nrcan 

    Examples
    --------
    rinex2snr mchn 2018 15  -archive sopac
        station mchn, year/doy 2022/15,sopac archive using GPS orbits

    rinex2snr mchn 2022 15  -archive sopac
        station mchn, year/doy 2022/15,sopac archive using multi-GNSS GFZ orbits

    rinex2snr mchn 2022 15  -archive sopac -orb gps
        station mchn, year/doy 2022/15,sopac archive using GPS orbits

    rinex2snr mchn 2022 15  -orb rapid -archive sopac
        now explicitly using rapid multi-GNSS orbits 

    rinex2snr mchn 2022 15  -orb rapid -archive sopac
        now explicitly using final multi-GNSS orbits (includes Beidou)

    rinex2snr mchn 2022 15  -orb rapid -archive sopac -overwrite T
        have an SNR file, but you want to make a new one

    rinex2snr p041 2022 15  -orb rapid -rate high -archive unavco
        now using high-rate data from unavco and multi-GNSS orbits

    rinex2snr p041 2022 15 -nolook T
        using your own data stored as p0410150.22o in the working directory 
        your RINEX o file may also be gzipped. I believe Hatanaka compressed is also allowed. 

    rinex2snr 940050 2021 31 -archive jp
        GSI archive in Japan - password required. Station names are six characters

    rinex2snr mchl00aus 2022 15  -orb rapid -archive ga 
        30 sec RINEX3 data for mchl00aus and Geoscience Australia

    rinex2snr mchl00aus 2022 15  -orb rapid -nolook T
        works if the RINEX 3 crx.gz or rnx files are in $REFL_CODE/2022/rinex/mchl

    rinex2snr mchl00aus 2022 15  -orb rapid -samplerate 30 -nolook T
        This should analyze a RINEX 3 file if it exists in your local working directory.
        it will not search anywhere else for the file.  It should be a 30 sec, 1 day file 
        for this example

    rinex2snr mchl00aus 2022 15  -orb rapid -samplerate 1 -nolook T -stream S -rate high
        This should analyze a RINEX 3 file if it exists in your local working directory.
        it will not search anywhere else for the file.  It should be a 1 sec, 1 day file 
        for this example with S being set for streaming in the filename.

    rinex2snr warn00deu 2023 87 -dec 5 -rate high -samplerate 1 -orb rapid -archive bkg-igs -stream S 
        1 sec data for warn00deu, 1 sec decimated to 5 sec, multi-GNSS, bkg IGS archive, streamed

    rinex2snr tgho  2019 1 -doy_end 365 -archive nz
        example for multiday SNR file creation

    Parameters
    ----------
    station : str
        4 or 9 character ID of the station, preferably lowercase
    year : int
        Year
    doy : int
        Day of year
    snr : int, optional
        SNR format. This tells the code what elevation angles to save data for. Will be the snr file ending.
        value options:

            66 (default) : saves all data with elevation angles less than 30 degrees

            99 : saves all data with elevation angles between 5 and 30 degrees

            88 : saves all data 
        
            50 : saves all data with elevation angles less than 10 degrees

    orb : str, optional
        Which orbit files to download. Value options:

            gps (default) : will use GPS broadcast orbit

            gps+glos : will use JAXA orbits which have GPS and Glonass (usually available in 48 hours)

            gnss : use GFZ final orbits, which is multi-GNSS (available in 3-4 days?), but from CDDIS archive

            gnss-gfz : GFZ orbits downloaded from GFZ instead of CDDIS, but do they include beidou?. Same as gnss3?

            nav : GPS broadcast, perfectly adequate for reflectometry. Same as gps.

            igs : IGS precise, GPS only

            igr : IGS rapid, GPS only

            jax : JAXA, GPS + Glonass, within a few days, missing block III GPS satellites

            gbm : GFZ Potsdam, multi-GNSS, not rapid, via CDDIS 

            grg : French group, GPS, Galileo and Glonass, not rapid

            esa : ESA, multi-GNSS

            gfr : GFZ rapid, GPS, Galileo and Glonass, since May 17 2021

            wum : Wuhan ultra-rapid, from CDDIS

            wum2 : Wuhan ultra-rapid, from Wuhan FTP

            rapid : GFZ rapid, multi-GNSS

            ultra: GFZ ultra-rapid, multi-GNSS

    rate : str, optional
        The data rate. Rather than numerical value, this tells the code which folder to use
        value options:

            low (default) : standard rate data. Usually 30 sec, but sometimes 15 sec.

            high : high-rate data

    dec : int, optional
        Decimation rate. 0 is default.

    nolook : bool, optional
        tells the code to retrieve RINEX files from your local machine. default is False

    archive : str, optional
        Select which archive to get the files from. Default is all
        value options:

            bev : (Austria Federal Office of Metrology and Surveying)

            bfg : (German Agency for water research, only Rinex 3, requires password)

            bkg-igs : IGS data at the BKG (German Agency for Cartography and Geodesy)

            bkg-euref : EUREF data at the BKG (German Agency for Cartography and Geodesy)

            cddis : (NASA's Archive of Space Geodesy Data)

            epn : Belgium

            ga : (Geoscience Australia)

            gfz : (GFZ, Germany)

            ignes : IGN in Spain, only RINEX 3

            jp : (GSI, Japan requires password)

            jeff : (My good friend Professor Freymueller!)

            ngs : (National Geodetic Survey, USA)

            nrcan : (Natural Resources Canada)

            nz : (GNS, New Zealand)

            sonel : (GLOSS archive for GNSS data)

            sopac : (Scripps Orbit and Permanent Array Center)

            special : (set aside files at UNAVCO for reflectometry users)

            unavco : (University Navstar Consortium, now Earthscope)

            all : (searches sopac, unavco, and sonel )

    doy_end : int, optional
        end day of year to be downloaded. 

    year_end : int, optional
        end year. 

    overwrite : bool, optional
        Make a new SNR file even if one already exists (overwrite existing file).
        Default is False.

    translator : str, optional
        hybrid (default) : uses a combination of python and fortran to translate the files.

        fortran : uses fortran to translate (requires the fortran translator executable to exist)

        python : uses python to translate. (Warning: This can be very slow)

    srate : int, optional
        sample rate for RINEX 3 files only. Default is 30.

    mk : bool, optional
        Default is False.
        Use True for uppercase station names and for the non-standard file structure preferred 
        by some users. Look at the function the_makan_option in rinex2snr.py for more information.  
        The general requirement is that your RINEX 2.11 file should be normal RINEX or 
        gzipped normal RINEX. This flag allows access to Hatanaka/compressed files 
        stored locally and in $REFL_CODE/YYYY/snr/ssss where YYYY is the year and 
        ssss is station name

    weekly : bool, optional
        Takes 1 out of every 7 days in the doy-doy_end range (one file per week) - used to save cpu time.
        Default is False.

    strip : bool, optional
        Reduces observables since the translator does not allow more than 25
        Default is False.

    screenstats: bool, optional
        if true, prints more information to the screen

    gzip: bool, optional
        default is true, SNR files are gzipped after creation.

    monthly : bool, optional
        default is false. snr files created every 30 days instead of every day

    par : int, optional
        default is NOne.  parallel processing, valid up to 10

    timeout : int, optional
        am seeing some timeouts when using highrate downloads and requests.
        adding this optional parameter to let you set the timeout value, but 
        it has not been implemented everywhere.  right now just the BKG


    """

    vers = 'gnssrefl version ' + str(g.version('gnssrefl'))
    #print('You are running ', vers)

    archive_list_rinex3 = ['unavco', 'epn','cddis', 'bev', 'bkg', 'ga', 'epn', 'bfg','sonel','all','unavco2','nrcan','gfz','ignes','gnet']
    archive_list = ['sopac', 'unavco', 'sonel',  'nz', 'ga', 'bkg', 'jeff',
                    'ngs', 'nrcan', 'special', 'bev', 'jp', 'all','unavco2','cddis','ngs_hourly']

    archive_list_no_parallel = ['sopac','cddis','jeff']

    if False:
        print('RINEX 3 archives \n', archive_list_rinex3)
        print('\n')
        print('RINEX 2.11 archives \n', archive_list)
        sys.exit()

    # make sure environment variables exist.  set to current directory if not
    g.check_environ_variables()
    xdir = os.environ['REFL_CODE']
    #
    if doy_end is None:
        doy_end = doy
    if year_end is None:
        year_end = year

    # make sure directories are there for orbit files and snr files
    # and for logs
    for y in range(year, year_end+1):
        tmp1, tmp2 = g.define_logdir(station,y,1)
        ann = g.make_nav_dirs(y)
        f1 = xdir + '/' + str(y) + '/snr/' + station[0:4]
        if not os.path.isdir(f1):
            print('make output directory for snr files in ', y)
            subprocess.call(['mkdir','-p',f1])

    # when multi-GNSS orbits are reliably available
    gfz_avail = 2021 + 137/365.25

    if orb is None:
        # you asked for default
        if ((year + doy/365.25) > gfz_avail):
            orb = 'rapid'
        else:
            orb = 'nav'

    ns = len(station)
    if (ns == 4) or (ns == 6) or (ns == 9):
        pass
    else:
        print('Illegal input - Station name must have 4 (RINEX 2), 6 (GSI), or 9 (RINEX 3) characters. Exiting.')
        sys.exit()

    if len(str(year)) != 4:
        print('Year must be four characters long. Exiting.', year)
        sys.exit()

    # currently allowed orbit types - shanghai removed 2020sep08
    #
    orbit_list = ['gps', 'gps+glo', 'gnss', 'nav', 'igs', 'igr', 'jax', 'gbm',
                  'grg', 'wum', 'wum2', 'gfr', 'esa', 'ultra', 'rapid', 'gnss2',
                  'nav-sopac', 'nav-esa', 'nav-cddis', 'gnss3', 'gnss-gfz']
    if orb not in orbit_list:
        print('You picked an orbit type I do not recognize. Here are the ones I allow')
        print(orbit_list)
        print('Exiting')
        sys.exit()

    # if you choose GPS, you get the nav message
    if orb == 'gps':
        orb = 'nav'

    if orb == 'rapid':
        orb = 'gfr'

    # if you choose GNSS, you get the GFZ sp3 file  (precise)
    # I think gbm should be changed to 'gnss3' though perhaps not here
    if orb == 'gnss':
        orb = 'gbm'
        print('Using GBM orbit archived at CDDIS')


    # get orbit from IGS
    if orb == 'gnss2':
        # this code wants year month day....
        year, month, day = g.ydoy2ymd(year, doy)
        filename, fdir, foundit = g.avoid_cddis(year, month, day)
        orb = 'gbm'
        if not foundit:
            print('You picked the backup multi-GNSS option.')
            print('I tried to get the file from IGN and failed. Exiting')
            sys.exit()
        else:
            print('found GFZ orbits at IGN - warning, only a single file at a time')

    # if you choose GPS+GLO, you get the JAXA sp3 file 
    if orb == 'gps+glo':
        orb = 'jax'

    # default is to use hybrid for RINEX translator - UNLESS You chose fortran

    # these are currently accepted
    translator_accepted = ['fortran', 'hybrid', 'python']
    if translator not in translator_accepted:
        print(f'translator option must be one of {translator_accepted}. Exiting.')
        sys.exit()

    # check that the fortran exe exist
    if translator == 'fortran':
        if orb == 'nav':
            snrexe = g.gpsSNR_version()
            if not os.path.isfile(snrexe):
                print('You have selected the fortran and GPS only options.')
                print('However, the fortran translator gpsSNR.e does not exist.')
                print('We are changing your choice to the hybrid translator option.')
                translator = 'hybrid'
        else:
            snrexe = g.gnssSNR_version()
            if not os.path.isfile(snrexe):
                print('You have selected the fortran and GNSS options.')
                print('However, the fortran translator gnssSNR.e has not been properly installed.')
                print('We are changing your choice to hybrid option.')
                translator = 'hybrid'

    rate = rate.lower()
    # default is set to low.  set to high for 1sec files so the user doesn't have to
    if samplerate == 1:
        rate = 'high'
    rate_accepted = ['low', 'high']

    if rate not in rate_accepted:
        print('rate not set to either "low" or "high". Exiting')
        sys.exit()

    bkg = 'EUREF' # just so the code doesn't crash later on when variable
    # is sent to rinex2snr
    if 'bkg' in archive:
        if (archive == 'bkg'):
            print('You have not specified which BKG archive you want.')
            print('You must select bkg-igs or bkg-euref.')
            sys.exit()
        if 'euref' in archive:
            bkg = 'EUREF'
        else:
            bkg = 'IGS'
        # change archive name back to original name
        archive = 'bkg'


    # adding spanish archive
    highrate_list = ['unavco', 'nrcan', 'ga','bkg','cddis','ignes','bkg-igs','bkg-euref','gnet']  
    if ns == 9:
        # rinex3
        if rate == 'high':
            if (archive not in highrate_list) and (nolook == False):
                print('You have chosen an archive not supported by the code.')
                print(highrate_list)
                sys.exit()
        else:
            if archive not in archive_list_rinex3:
                print('You have chosen an archive not supported by the code.')
                print(archive_list_rinex3)
                sys.exit()
    else:
        # rinex2
        if rate == 'high':
            if archive not in highrate_list:
                if nolook:
                    print('You have chosen nolook, so I will proceed assuming you have the RINEX file.')
                    # change to lowrate since the code only uses low vs high for retrieving files from
                    # an archive
                    rate = 'low'
                else:
                    print('You have chosen highrate option.  But I do not support this archive: ',archive)
                    sys.exit()
        else:
            if archive not in archive_list:
                print('You picked an archive that is not allowed. Exiting')
                print(archive_list)
                sys.exit()

    skipit = 1
    if weekly:
        print('You have invoked the weekly option')
        skipit = 7
        print('Monthly and Weekly functions are not currently working. Resubmit.')
        sys.exit()
    if monthly:
        print('You have invoked the monthly option')
        skipit = 30
        print('Monthly and Weekly functions are not currently working. Resubmit.')
        sys.exit()

    # the Makan option
    if mk:
        print('You have invoked the Makan option')

    if stream not in ['R', 'S']:
        stream = 'R'

    args = {'station': station, 'year':year, 'doy':doy, 'isnr': snr, 'orbtype': orb, 'rate': rate, 
            'dec_rate': dec, 'archive': archive, 'nol': nolook, 'overwrite': overwrite, 
            'translator': translator, 'srate': samplerate, 'mk': mk, 'stream': stream, 
            'strip': strip, 'bkg': bkg, 'screenstats': screenstats, 'gzip' : gzip, 'timeout' : timeout }
    MJD1 = int(g.ydoy2mjd(year,doy))
    MJD2 = int(g.ydoy2mjd(year_end,doy_end))


    # queue which handles any exceptions any of the processes encounter
    manager = multiprocessing.Manager()
    error_queue = manager.Queue()


    if MJD1 == MJD2:
        print('Only one day being analyzed, parallel processing turned off')
        par = None
    if archive in archive_list_no_parallel:
        print('You have chosen an archive that is unfriendly to multiple simultaneous download')
        print('requests. Your request for parallel processing has been declined.')
        par = None

    if not par:
        print('No parallel processing')
        #mjd_list = range(MJD1, MJD2+1, skipit)
        mjd_list = {}; mjd_list[0] = [MJD1, MJD2]
        s1 = time.time()
        process_jobs_multi(index=0,args=args,datelist=mjd_list,error_queue=error_queue)
        s2 = time.time()
        print('That took ', round(s2-s1,2), ' seconds')

    else:
        print('You have chosen parallel processing')
        if par > 10:
            print('For now we will only allow ten simultaneous processes. Submit again. Exiting.')
            sys.exit()
        else:
            numproc = par
            # get a list of times in MJD associated with the multiple spawned processes
            # this does not work for skipping dates though ... 

            datelist, numproc = guts2.make_parallel_proc_lists_mjd(year, doy, year_end, doy_end, numproc)
            print(datelist)

            # make a list of process IDs
            index = list(range(numproc))

            s1 = time.time()

            pool = multiprocessing.Pool(processes=numproc)

            partial_process = partial(process_jobs_multi, args=args,datelist=datelist,error_queue = error_queue)

            pool.map(partial_process,index)

            pool.close()
            pool.join()
            s2 = time.time()
            print('That took ', round(s2-s1,2), ' seconds')

def process_jobs_multi(index,args,datelist,error_queue):
    """
    runs the rinex2snr queue

    Parameters
    ==========
    index : int
        which job to run
        args : dict
        dictionary of parameters for run_rinex2snr
    datelist: dict
        start and stop dates in MJD
    error_queue:? 
        not sure how to describe this

    """

    # should try be on each submission?  or each list of submissions?
    try:
        d1 = datelist[index][0]; d2 = datelist[index][1]
        mjd_list = list(range(d1, d2+1))
        for mjd in mjd_list:
            y, d = g.modjul_to_ydoy(mjd)
            args['year'] = y
            args['doy'] = d
            rnx.run_rinex2snr(**args)

    except Exception as e:
        print('Error of some kind processing year/doy ', y, d)
        error_queue.put(e)

    return

def process_jobs(mjd_list, args):
    """
    this is not being used - calls should be sent to function above instead
    """
    for mjd in mjd_list:
        y, d = g.modjul_to_ydoy(mjd)
        args['year'] = y
        args['doy'] = d
        rnx.run_rinex2snr(**args)

    return

def main():
    args = parse_arguments()
    rinex2snr(**args)


if __name__ == "__main__":
    main()





import argparse
import gnssrefl.gps as g

def main():
    """
    converts MJD to year, month, day, hour, minute, second
    and prints that to the screen

    Parameters
    ----------
    mjd : float
        modified julian date

    """

    parser = argparse.ArgumentParser()
    parser.add_argument("mjd", help="Modified Julian Date", type=float)

    args = parser.parse_args()
    the_mjd= args.mjd

    filler = g.mjd_to_datetime(the_mjd)
    print(filler)

if __name__ == "__main__":
    main()



import subprocess
from unittest import mock

import os
import urllib.error
import wget

from gnssrefl.gps import *


def test_get_sopac_navfile_working(mocker):
    mocker.patch("wget.download")
    mocker.patch("os.path.exists")
    mocker.patch("subprocess.call")
    os.path.exists.return_value = True
    assert (
        get_sopac_navfile("auto1050.20n", "2020", "20", "105")
        == "auto1050.20n"
    )
    wget.download.assert_called_once_with(
        "ftp://garner.ucsd.edu/pub/rinex/2020/105/auto1050.20n.Z",
        "auto1050.20n.Z",
    )
    os.path.exists.assert_called_once_with("auto1050.20n")
    subprocess.call.assert_called_once_with(["uncompress", "auto1050.20n.Z"])


# this should be changed to run on january 1, 2021 which probably 
# has a corrupt file on the sopac database
def test_get_sopac_navfile_error(mocker):
    mocker.patch("wget.download")
    mocker.patch("os.path.exists")
    mocker.patch("subprocess.call")
    wget.download.side_effect = urllib.error.URLError("404")
    os.path.exists.return_value = False
    assert (
        get_sopac_navfile("p1031050.20.snr66", "2020", "20", "105")
        == "p1031050.20.snr66"  # should this return None instead?
    )
    wget.download.assert_called_once_with(
        "ftp://garner.ucsd.edu/pub/rinex/2020/105/p1031050.20.snr66.Z",
        "p1031050.20.snr66.Z",
    )
    os.path.exists.assert_called_once_with("p1031050.20.snr66")
    assert subprocess.call.mock_calls == [
        mock.call(["rm", "-f", "p1031050.20.snr66.Z"]),
        mock.call(["rm", "-f", "p1031050.20.snr66"]),
    ]



# -*- coding: utf-8 -*-
import argparse
import os
import subprocess
import sys

import gnssrefl.gps as g

def main():
    """
    Decimates and strips out SNR data from a RINEX 2.11 file. 
    Only RINEX, no Hatanaka RINEX or gzipped files allowed. If you would like to add
    these features, please submit a PR.

    Examples
    --------

    smoosh p0410010.22o 5
        decimates a highrate rinex 2.11 file to 5 seconds

    smoosh p0410010.22o 5 -snr T
        also eliminates all observation types except for SNR

    Parameters
    ----------
    rinex : str
        rinex 2.11 filename
    dec : int
        decimation value in seconds
    snr : bool, optional
        whether you want only SNR observables
        helpful when you have too many observables in your file.

    """

    parser = argparse.ArgumentParser()
    parser.add_argument("rinexfile", help="filename, rinex 2 only", type=str)
    parser.add_argument("dec", help="decimation value, seconds", type=int)
    parser.add_argument("-snr", help="whether you want SNR data stripped out", type=str,default=None)

    args = parser.parse_args()
    rinexfile = args.rinexfile
    dec = args.dec 
    gexe = g.gfz_version()
    snr = True
    rinexfile2 = rinexfile + '.tmp'

    if not os.path.exists(gexe):
        print('Required gfzrnx executable does not exist. Exiting.')
        sys.exit()


    if not os.path.isfile(rinexfile):
        print('Your RINEX input file does exist:', rinexfile)
        sys.exit()

    crate = str(dec)
    iall = 'GRE'
    sig = 'S1,S2C,S2,S5,S6,S7'
    subprocess.call([gexe,'-finp', rinexfile, '-fout', rinexfile2, '-vo','2','-ot', sig, '-smp', crate, '-satsys',iall,'-f'])
    # move new file to original name
    subprocess.call(['rm','-f', rinexfile])
    subprocess.call(['mv','-f', rinexfile2, rinexfile])


if __name__ == "__main__":
    main()



import argparse
import wget
import sys
import os
import gnssrefl.gps as g


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("station", help="station name", type=str)
    args = parser.parse_args().__dict__

    # only return a dictionary of arguments that were added from the user - all other defaults will be set in code below
    return {key: value for key, value in args.items() if value is not None}


def download_unr(station: str):
    """
    Command line interface for downloading time series from the 
    University of Nevada Reno website

    This code is not actively maintained.

    Examples
    --------
    download_unr p041

    download_unr sc02

    Parameters
    ----------
    station : str
        4 character ID of the station name

    """

    if len(station) != 4:
        print('illegal station name-must be 4 characters')
        sys.exit()
    station = station.upper()
    url= 'http://geodesy.unr.edu/gps_timeseries/tenv3/IGS14/'
    fname = station + '.tenv3'
    stationL = station.lower() # lower case
    url = url + fname
    # file will be stored in this directory
    xdir = os.environ['REFL_CODE']
    outdir = xdir  + '/Files/'
    if not os.path.exists(outdir) :
        subprocess.call(['mkdir', outdir])

    g.check_environ_variables()

    # store in Files subdirectory
    myfname = xdir + '/Files/' + stationL + '_igs14.tenv3'
    try:
        wget.download(url, out=myfname)
    except:
        print('\n download failed:', url)

    if os.path.exists(myfname):
        print('\n SUCCESS:', myfname)


def main():
    args = parse_arguments()
    download_unr(**args)


if __name__ == "__main__":
    main()



# codes for subdaily module. primarily for tidal applications
import argparse
import datetime
import json
import matplotlib.pyplot as plt
import numpy as np
import os
import sys
import warnings

from datetime import date

# support code
import gnssrefl.gnssir_v2 as guts2
import gnssrefl.gps as g

import gnssrefl.sd_libs as sd


import scipy
import scipy.interpolate as interpolate
from scipy.interpolate import interp1d
import math

#KL trying to move plots and output files to sd_libs.py


def output_names(txtdir, txtfile,csvfile,jsonfile):
    """
    figures out what the names of the outputs are going to be

    I have modified this so it always returns plain txt. csv will
    simply be written out in addition.

    this function no longer has much point. 

    Parameters
    ----------
    txtdir : str
        the directory where the results should be written out
    txtfile : str
        name of the output file 
    csvfile : bool
        cl input whether the output file should be csv format
    jsonfile : bool
        cl input for whether the output file should be in the json format

    Returns
    -------
    writetxt : bool
        whether output should be plain txt
    writecsv : bool
        whether output should be csv format
    writejson : bool
        whether output should be json format
    outfile : str
        output filename

    """
    writetxt = True
    #if txtfile == None:
    #    writetxt = False

    #writecsv = True
    #if csvfile == None:
    #    writecsv = False
    writecsv = csvfile

    # no longer allowed
    #writejson = True
    #if writejson == None:
    #    writejson = False
    writejson = False 
    #if writejson:
    #    outfile = txtdir + '/' + jsonfile
    #if writecsv:
    #    outfile = txtdir + '/' + csvfile
    if writetxt:
        outfile = txtdir + '/' + txtfile
    #if (writecsv) and (writetxt) :
    #    print('You cannot simultaneously write out a csvfile and a txtfile')
    #    print('Default to writing only a txtfile')
    #    writecsv = False

    print('Main plain txt outputfile ', outfile)
    if writecsv:
        print('csv file will also be written')

    return writetxt,writecsv,writejson,outfile

def write_subdaily(outfile,station,ntv,csv,extraline,**kwargs):
    """
    writes out the subdaily results. currently only works for plain txt

    >> this code should be moved to the library

    Parameters
    ----------
    input : str
        output filename
    station : str
        4 character station name, lowercase
    nvt : numpy multi-dimensional
        the variable with the LSP results read via np.loadtxt
    csv : bool
        whether both csv and txt file should be written 
    extraline: bool
        whether the header has an extra line

    """
    RHdot_corr= kwargs.get('RHdot_corr',[])

    newRH = kwargs.get('newRH',[])

    newRH_IF = kwargs.get('newRH_IF',[])

    original = False
    if len(RHdot_corr) + len(newRH) + len(newRH_IF) == 0:
        original = True
        #print('LSP series being written')
    # 
    write_IF_corrected = False
    if len(newRH_IF) > 0:
        print('\nIF corrected values being written - make sure you use the correct column! \n')
        write_IF_corrected = True
    extra = False
    if len(RHdot_corr) > 0:
        print('\nRHdot corrected values being written - make sure you use the correct column \n')
        extra = True

    N= len(ntv)
    nr,nc = ntv.shape
    if nr == 0:
        print('No results in this file, so nothing to write out.')
        return
    print(nr, ' observations will be written to:', outfile)

    N= nr
    # the lucky people get plain txt AND csv
    if csv:
        # exchange txt for csv
        if original:
            csv_outfile = outfile[0:-3] + 'csv'
        if extra: # or just add the csv.  
            csv_outfile = outfile + '.csv'
        if write_IF_corrected: # or just add the csv.  
            csv_outfile = outfile + '.csv'

        print('Opening this as well ', csv_outfile)
        fout_csv = open(csv_outfile, 'w+')

    # everyone gets a plain txt file
    fout = open(outfile, 'w+')

    if extra:
        write_out_header(fout,station,extraline,extra_columns=True)
        if csv:
            write_out_header(fout_csv,station,extraline,extra_columns=True)

    else:
        if write_IF_corrected:
            write_out_header(fout,station,extraline, IF=True)
            if csv:
                write_out_header(fout_csv,station,extraline, IF=True)
        else:
            write_out_header(fout,station,extraline)
            if csv:
                write_out_header(fout_csv,station,extraline)

    dtime = False
    for i in np.arange(0,N,1):
        year = int(ntv[i,0]); doy = int(ntv[i,1])
        year, month, day, cyyyy,cdoy, YMD = g.ydoy2useful(year,doy)
        rh = ntv[i,2]; UTCtime = ntv[i,4]; 
        dob, year, month, day, hour, minute, second = g.ymd_hhmmss(year,doy,UTCtime,dtime)
        #ctime = g.nicerTime(UTCtime); 
        # taking hte lazy approach - if someone wants csv write out txt and csv
        if csv:
            if original:
                fout_csv.write(" {0:4.0f},{1:3.0f},{2:7.3f},{3:3.0f},{4:6.3f},{5:6.2f},{6:6.2f},{7:6.2f},{8:6.2f},{9:4.0f},{10:3.0f},{11:2.0f},{12:8.5f},{13:6.2f},{14:7.2f},{15:12.6f},{16:1.0f},{17:2.0f},{18:2.0f},{19:2.0f},{20:2.0f},{21:2.0f} \n".format(year, doy, rh,ntv[i,3],UTCtime,ntv[i,5],ntv[i,6],ntv[i,7],ntv[i,8], ntv[i,9], ntv[i,10],ntv[i,11], ntv[i,12],ntv[i,13], ntv[i,14], ntv[i,15], ntv[i,16],month,day,hour,minute, int(second)))

            if extra: 
                fout_csv.write(" {0:4.0f},{1:3.0f},{2:7.3f},{3:3.0f},{4:6.3f},{5:6.2f},{6:6.2f},{7:6.2f},{8:6.2f},{9:4.0f},{10:3.0f},{11:2.0f},{12:8.5f},{13:6.2f},{14:7.2f},{15:12.6f},{16:1.0f},{17:2.0f},{18:2.0f},{19:2.0f},{20:2.0f},{21:2.0f},{22:10.3f},{23:10.3f} \n".format(year, doy, rh,ntv[i,3],UTCtime,ntv[i,5],ntv[i,6],ntv[i,7],ntv[i,8], ntv[i,9], ntv[i,10],ntv[i,11], ntv[i,12],ntv[i,13], ntv[i,14], ntv[i,15], ntv[i,16],month,day,hour,minute, int(second), newRH[i],RHdot_corr[i]))
            if write_IF_corrected:
                fout_csv.write(" {0:4.0f},{1:3.0f},{2:7.3f},{3:3.0f},{4:6.3f},{5:6.2f},{6:6.2f},{7:6.2f},{8:6.2f},{9:4.0f},{10:3.0f},{11:2.0f},{12:8.5f},{13:6.2f},{14:7.2f},{15:12.6f},{16:1.0f},{17:2.0f},{18:2.0f},{19:2.0f},{20:2.0f},{21:2.0f},{22:10.3f},{23:10.3f},{24:10.3f},\n".format(year, doy, rh,ntv[i,3], 
                    UTCtime,ntv[i,5],ntv[i,6],ntv[i,7],ntv[i,8], ntv[i,9], ntv[i,10],ntv[i,11], 
                    ntv[i,12],ntv[i,13], ntv[i,14], ntv[i,15], ntv[i,16],month,day, hour,minute, 
                    int(second), ntv[i,22], ntv[i,23], newRH_IF[i] ))

        # just so i don't have to re-indent everything
        if True:
            if write_IF_corrected:
                fout.write(" {0:4.0f} {1:3.0f} {2:7.3f} {3:3.0f} {4:6.3f} {5:6.2f} {6:6.2f} {7:6.2f} {8:6.2f} {9:4.0f} {10:3.0f} {11:2.0f} {12:8.5f} {13:6.2f} {14:7.2f} {15:12.6f} {16:1.0f} {17:2.0f} {18:2.0f} {19:2.0f} {20:2.0f} {21:2.0f} {22:10.3f} {23:10.3f} {24:10.3f} \n".format(year, doy, rh,ntv[i,3], 
                    UTCtime,ntv[i,5],ntv[i,6],ntv[i,7],ntv[i,8], ntv[i,9], ntv[i,10],ntv[i,11], 
                    ntv[i,12],ntv[i,13], ntv[i,14], ntv[i,15], ntv[i,16],month,day, hour,minute, 
                    int(second), ntv[i,22], ntv[i,23], newRH_IF[i] ))

            if extra:
                    fout.write(" {0:4.0f} {1:3.0f} {2:7.3f} {3:3.0f} {4:6.3f} {5:6.2f} {6:6.2f} {7:6.2f} {8:6.2f} {9:4.0f} {10:3.0f} {11:2.0f} {12:8.5f} {13:6.2f} {14:7.2f} {15:12.6f} {16:1.0f} {17:2.0f} {18:2.0f} {19:2.0f} {20:2.0f} {21:2.0f} {22:10.3f} {23:10.3f} \n".format(year, doy, rh, ntv[i,3], 
                        UTCtime,ntv[i,5],ntv[i,6],ntv[i,7],ntv[i,8], ntv[i,9], ntv[i,10],ntv[i,11], ntv[i,12],ntv[i,13], ntv[i,14], 
                        ntv[i,15], ntv[i,16],month,day,hour,minute, int(second), newRH[i], RHdot_corr[i]))

            if original:
                fout.write(" {0:4.0f} {1:3.0f} {2:7.3f} {3:3.0f} {4:6.3f} {5:6.2f} {6:6.2f} {7:6.2f} {8:6.2f} {9:4.0f} {10:3.0f} {11:2.0f} {12:8.5f} {13:6.2f} {14:7.2f} {15:12.6f} {16:1.0f} {17:2.0f} {18:2.0f} {19:2.0f} {20:2.0f} {21:2.0f} \n".format(year, doy, rh,ntv[i,3],UTCtime,ntv[i,5],
                    ntv[i,6],ntv[i,7],ntv[i,8], ntv[i,9], ntv[i,10],ntv[i,11], ntv[i,12],ntv[i,13], ntv[i,14], 
                    ntv[i,15], ntv[i,16],month,day,hour,minute, int(second)))

    # close both files if necessary
    if csv:
        fout_csv.close()

    fout.close()


def readin_and_plot(station, year,d1,d2,plt2screen,extension,sigma,writecsv,azim1,azim2,ampl,
        peak2noise,txtfile,h1,h2,kplt,txtdir,default_usage,hires_figs,fs,**kwargs):
    """
    Reads in RH results and makes various plots to help users assess the quality of the solution

    This is basically "section 1" of the code

    Parameters
    ----------
    station : str
        4 character station name
    year : int
        full year
    d1 : int
        first day of year evaluated
    d2 : int 
        last day of year evaluated
    plt2screen : bool
        if True plots are displayed to the screen
    extension : str
        allow user to specify an extension for results (i.e. gnssir was run using extension string)
    sigma : float
         how many standard deviations away from mean you allow for the crude outlier detector.  
    writecsv : bool
         whether output is written in csv format. 
    azim1 : float
        minimum azimuth value (degrees)
    azim2 : float
        maximum azimuth value (degrees)
    ampl : float
        minimum LSP amplitude allowed
    peak2noise : float
        minim peak2noise value to set solution good
    txtfile : str
        name of plain text output file
    h1 : float
        minimum reflector height (m)
    h2 : float
        maximum reflector height (m)
    kplt : bool
        special plot made 
    txtdir : str
        directory where the results will be written
    default_usage : bool
        flag as to whether you are using this code for subdaily or for rh_plot.
        this changes the plots a bit.
    hires_figs: bool
        whether to switch from png to eps
    fs : int
        fontsize for figure axes

    Returns
    -------
    tv : numpy array
        LSP results (augmented)
    otimes : datetime object 
        times of observations 
    fname : str
        initial result file - colated
    fname_new : str
        result file with outliers removed

    """

    # felipe nievinski alternate definition
    alt_sigma = kwargs.get('alt_sigma', False)

    # used different name here
    csv2= writecsv

    xdir = os.environ['REFL_CODE']
    print('Will remove daily outliers greater than ', sigma, ' sigma')
    if not os.path.exists(txtdir):
        os.makedirs(txtdir)
    print('Plot to the screen has been set to ', plt2screen)

    print('>>>>>>>>>>>>>>>>>>>>>>>> readin RH data <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<')
    if txtfile == '':
        print('Read in the RH retrievals for ', year, ' and these days: ',d1,d2)
        if len(extension) > 0:
            print('Using the results in the ', extension , ' subdirectory.')
        if (d2 < d1):
            print('First day of year must be less than last day of year. Exiting')
            sys.exit()
    # where the LSP results are kept
        direc = xdir + '/' + str(year) + '/results/' + station  + '/' + extension + '/'
    # datetime object for time
        obstimes = []
        tv = np.empty(shape=[0, 17])
        if os.path.isdir(direc):
            all_files = os.listdir(direc)
            for f in all_files:
            # only evaluate txt files
                if (len(f) ==  7) and (f[4:7] == 'txt'):
                    day = int(f[0:3])
                    if (day >= d1) & (day <= d2):
                        fname = direc + f
                        try:
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore")
                                a = np.loadtxt(fname,comments='%')
                            if len(a) > 0:
                                tv = np.append(tv, a,axis=0)
                        except:
                            print('some issue with ',fname)

    else:
        print('using external file of concatenated results', txtfile)
        tv = np.loadtxt(txtfile,comments='%')

     
    tv,t,rh,fdoy,ldoy= apply_new_constraints(tv,azim1,azim2,ampl,peak2noise,d1,d2,h1,h2)

    tvoriginal = tv
    nr,nc = tvoriginal.shape
    print('RH retrievals after all commandline constraints', nr)

    # now that you have everything the way you want it .... 
    # make the datatime objects
    nr,nc = tv.shape
    otimes = []
    for ijk in range(0,nr):
        dtime, iyear,imon,iday,ihour,imin,isec = g.ymd_hhmmss(tv[ijk,0],tv[ijk,1],tv[ijk,4],True)
        otimes.append(dtime)

    # make arrays to save number of RH retrievals on each day
    residuals = np.empty(shape=[0,1])
    real_residuals = np.empty(shape=[0,1])
    nval = []; tval = []; y = tv[0,0];  
    # constellation spec number of values
    Cval=[]; Gval =[]; Rval=[]; Eval=[]
    stats = np.empty(shape=[0,3])
    # only look at the doy range where i have data
    for d in range(fdoy,(ldoy+1)):
        ii = (tv[:,1] == d) ; tmp = tv[ii,:]
        dtime, iyear,imon,iday,ihour,imin,isec = g.ymd_hhmmss(year,d,12,True)
        tval.append(dtime)
        n = len(tv[ii,1])
        # total
        nval.append(n)
#       https://stackoverflow.com/questions/26786946/how-to-return-indices-of-values-between-two-numbers-in-numpy-array
        gi =  (tmp[:,10] < 100);
        ri =  (tmp[:,10] > 100) * (tmp[:,10] < 200);
        ei =  (tmp[:,10] > 200) * (tmp[:,10] < 300);
        ci =  (tmp[:,10] > 300); # beidou

        #print( len(tmp[ri,1])) print( len(tmp[ei,1]))

        if (n > 0):
            if alt_sigma:
            # make sigma clipping more robust against outliers:
            # https://en.wikipedia.org/wiki/Median_absolute_deviation#Relation_to_standard_deviation
                rhavg = np.median(tv[ii,2]); 
                rhstd = np.median(abs(tv[ii,2]-rhavg))/0.6745;            
            else:
                rhavg = np.mean(tv[ii,2]); 
                rhstd = np.std(tv[ii,2]); 

            newl = [dtime, rhavg, rhstd]
            stats = np.append(stats, [newl], axis=0)
            if rhstd == 0:
                rhstd = 1
                
            b = ( tv[ii,2] - rhavg*np.ones( len(tv[ii,2]) ))/ rhstd
            bb =  tv[ii,2] - rhavg*np.ones( len(tv[ii,2]) )
            
            residuals = np.append(residuals, b)
            real_residuals = np.append(real_residuals, bb)
            Gval = np.append(Gval, len(tmp[gi,1]))
            Eval = np.append(Eval, len(tmp[ei,1]))
            Rval = np.append(Rval, len(tmp[ri,1]))
            Cval = np.append(Cval, len(tmp[ci,1]))
        else:
            Gval = np.append(Gval, 0)
            Eval = np.append(Eval, 0)
            Rval = np.append(Rval, 0)
            Cval = np.append(Cval, 0)

    ii = (np.absolute(residuals) > sigma) # data you like
    jj = (np.absolute(residuals) < sigma) # data you do not like ;-)

    # I think this is plots for section I
    if plt2screen:

        minAz = float(np.min(tv[:,5])) ; maxAz = float(np.max(tv[:,5]))

        if default_usage:
            sd.numsats_plot(station,tval,nval,Gval,Rval,Eval,Cval,txtdir,fs,hires_figs,year)
            # was testing some things out
            #testing_nvals(Gval, Rval, Eval, Cval)

            sd.two_stacked_plots(otimes,tv,station,txtdir,year,d1,d2,hires_figs)
            sd.stack_two_more(otimes,tv,ii,jj,stats, station, txtdir,sigma,kplt,hires_figs,year)
            sd.print_badpoints(tv[ii,:],residuals[ii],txtdir,real_residuals[ii])
        else:
            # make both plots cause life is short
            # I do not think this is currently used ... 
            sd.rh_plots(otimes,tv,station,txtdir,year,d1,d2,True)
            sd.rh_plots(otimes,tv,station,txtdir,year,d1,d2,False)

        #plt.show()


    # now write things out - using txtdir variable sent 
    fname =     txtdir  + '/' + station + '_' + str(year) + '_subdaily_all.txt'
    fname_new = txtdir  + '/' + station + '_' + str(year) + '_subdaily_edit.txt'

    extraline = ''

    editedtv = tv[jj,:]
    nr,nc = editedtv.shape

    # write out the initial timeseries which are basically the concatenated values for multiple days
    write_subdaily(fname,station,tvoriginal,writecsv,extraline)

    if default_usage:
        print('Edited observations',nr)
        extraline = 'outliers removed/restrictions'
        write_subdaily(fname_new,station,editedtv,writecsv,extraline)
        NV = len(tvoriginal)
        print('Percent of observations removed: ', round(100*(NV-nr)/NV,2))
    
    # now return the names of the output files ... 

    return tv,otimes, fname, fname_new


def fract_to_obstimes(spl_x):
    """
    this does not seem to be used

    Parameters
    ----------
    spl_x : numpy array
        fractional time 

    obstimes : numpy array
        datetime format
    """
    N=len(spl_x)
    obstimes = np.empty(shape=[0, 1])
    year = np.floor(spl_x)
    fdoy = 365.25*(spl_x - year)
    doy = np.floor(fdoy)
    fhours = 24*(fdoy - doy)
    hours = np.floor(fhours)
    leftover = 60*(fhours - hours)
    minutes = np.floor(leftover)
    for i in range(0,N):
        y = int(year[i]); d = int(doy[i])
        y, month, day, cyyyy,cdoy, YMD = g.ydoy2useful(y,d)
        h= int(hours[i])
        m = int(minutes[i])
        #print(y,month,day,h,m)
        obstimes = np.append(obstimes, datetime.datetime(year=y, month=month, day=day, hour=h, minute=m, second=0 ))

    return obstimes

def spline_in_out(x,y,knots_per_day):
    """

    Parameters
    ----------
    x : numpy of floats
        time of observations in fractional days

    y : numpy of floats
        reflector heights in meters

    knots_per_day : int
        number of knots per day

    Returns
    -------
    xx : numpy of floats
        regularly spaced observations

    spline(xx): numpy of floats
        spline value at those times

    """
    Ndays = round(x.max()-x.min())
    numKnots = int(knots_per_day*(Ndays))
    #print('xmin, xmax',x.min(), x.max(), 'knots', numKnots,Ndays )
    x1 = x.min()+0.1/365.25
    x2 = x.max()-0.1/365.25
    knots =np.linspace(x1,x2,num=numKnots)
    t, c, k = interpolate.splrep(x, y, s=0, k=3,t=knots,task=-1)
#   calculate water level hourly for now
    N = int(Ndays*24 )
    xx = np.linspace(x.min(), x.max(), N)
    spline = interpolate.BSpline(t, c, k, extrapolate=False)

    return xx,spline(xx) 
    
def write_out_header(fout,station,extraline,**kwargs):
    """
    writes out header for results file ... 

    Parameters
    ----------
    fout : fileID

    station : str
        4 character station name

    extraline : bool
        not sure why this is here

    """
    extra_columns = kwargs.get('extra_columns',False)
    IFcol = kwargs.get('IF',False)
    vers = 'gnssrefl v' + str(g.version('gnssrefl')) 
    xxx = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M")) 
    #d= 'Results for ' + station + 'calculated with' + vers + ' on '  + xxx
    d=f'Results for {station} calculated with {vers} on {xxx}'
    fout.write("% {0:40s} \n".format(d))
    #fout.write('% gnssrefl, https://github.com/kristinemlarson \n')
    if len(extraline) > 0:
        fout.write('% IMPORTANT {0:s} \n'.format(  extraline ))
    fout.write('% Traditional phase center corrections have NOT been applied \n')
    if extra_columns:
        fout.write("% (1)  (2)   (3) (4)  (5)     (6)   (7)    (8)    (9)   (10)  (11) (12) (13)    (14)     (15)    (16) (17) (18,19,20,21,22)    (23)        (24) \n")
        fout.write("% year, doy, RH, sat,UTCtime, Azim, Amp,  eminO, emaxO,NumbOf,freq,rise,EdotF, PkNoise, DelT,    MJD, refr, MM,DD,HH,MM,SS,  RH with,     RHdot \n")
        fout.write("%             m       hours   deg   v/v    deg   deg                1/-1       ratio    minute        1/0                    RHdotCorr    Corr m    \n")
    else:
        if IFcol:
            fout.write('% Reflector heights set to the L1-phase center are in column 25 \n')
            fout.write("% (1)  (2)   (3) (4)  (5)     (6)   (7)    (8)    (9)   (10)  (11) (12) (13)    (14)     (15)    (16) (17) (18,19,20,21,22)   (23)      (24)    (25)\n")
            fout.write("% year, doy, RH, sat,UTCtime, Azim, Amp,  eminO, emaxO,NumbOf,freq,rise,EdotF, PkNoise, DelT,    MJD, refr, MM,DD,HH,MM,SS, RH with,   RHdot,    RH with  \n")
            fout.write("%             m       hours   deg   v/v    deg   deg                1/-1       ratio    minute        1/0                   RHdotCorr  Corr m    IF Corr  m  \n")
        else:
            fout.write("% (1)  (2)   (3) (4)  (5)     (6)   (7)    (8)    (9)   (10)  (11) (12) (13)    (14)     (15)    (16) (17) (18,19,20,21,22)\n")
            fout.write("% year, doy, RH, sat,UTCtime, Azim, Amp,  eminO, emaxO,NumbOf,freq,rise,EdotF, PkNoise, DelT,    MJD, refr, MM,DD,HH,MM,SS \n")
            fout.write("%             m       hours   deg   v/v    deg   deg                1/-1       ratio    minute        1/0                  \n")



def apply_new_constraints(tv,azim1,azim2,ampl,peak2noise,d1,d2,h1,h2):
    """
    cleaning up the main code. this sorts data and applies various "commandline" constraints
    tv is the full set of results from gnssrefl

    Parameters
    ----------
    tv : numpy array
        lsp results
    azim1 : float
        min azimuth (deg)
    azim2 : float
        max azimuth (deg)
    ampl : float
        required amplitude for periodogram
    peak2noise : float
        require peak2noise criterion
    d1 : int
        min day of year
    d2 : int
        max day of year
    h1 : float
        min reflector height (m)
    h2 : float
        max reflector height (m)

    Returns
    -------
    tv : numpy array
        edited from input
    t : numpy of floats
        crude time for obs, in fractional days
    rh : numpy of floats
        reflector heights (m)
    firstdoy : int
        first day of year
    lastdoy : int 
        last day of year
    """
    nr,nc=tv.shape
    tvoriginal = tv
    print('Number of initial RH retrievals', nr)
    nroriginal = nr
    if (nr == 0):
        print('Exiting')
        sys.exit()

    ii = (tv[:,2] <= h2) & (tv[:,2] >= h1)
    tv = tv[ii,:]

    if ((nr-len(tv)) > 0):
        print(nr-len(tv) , ' points removed for reflector height constraints ',h1,h2, ' (m)')

    # not sure these need to be here - are they used elsewhere?
    # could be moved to apply_new_constraints
    
    t=tv[:,0] + (tv[:,1] + tv[:,4]/24)/365.25
    rh = tv[:,2]

# sort the data
    ii = np.argsort(t)
    t = t[ii] ; rh = rh[ii]
    # store the sorted data
    tv = tv[ii,:]

    # apply azimuth constraints
    nr,nc = tv.shape
    ii = (tv[:,5]  >= azim1) & (tv[:,5] <= azim2)
    tv = tv[ii,:]
    print(nr-len(tv) , ' points removed for azimuth constraints ',azim1,azim2)

    # now apply amplitude constraint
    nr,nc = tv.shape
    ii = (tv[:,6]  >= ampl) ; tv = tv[ii,:]
    print(nr-len(tv) , ' points removed for amplitude constraint ')

    # now apply peak2noise constraint
    nr,nc = tv.shape
    ii = (tv[:,13]  >= peak2noise) ; tv = tv[ii,:]
    print(nr-len(tv) , ' points removed for peak2noise constraints ')


    # silly - why read it if you are not going to use it
    # and restrict by doy - mostly to make testing go faster
    ii = (tv[:,1] >= d1) & (tv[:,1] <= d2)
    tv = tv[ii,:]
    firstdoy = int(min(tv[:,1]))
    lastdoy =  int(max(tv[:,1]))

    return tv,t,rh,firstdoy,lastdoy


def flipit(tvd,col):
    """
    take RH values from the first and last day and attaches
    them as fake data to make the spline fit stable.  
    Also fill the temporal gaps with fake data

    Parameters
    ----------
    tvd : numpy array of floats
        output of LSP runs. 
    col : integer
        column number (in normal speak) of the RH results
        in python-speak, this has one subtracted

    Returns
    -------
    tnew : numpy array of floats
        time in days of year
    ynew : numpy array
        RH in meters 

    """
    nr,nc = np.shape(tvd)
    print(nr,nc)
    # sort it just to make sure ...
    tnew = tvd[:,1] + tvd[:,4]/24
    # change from normal columns to python columns
    ynew = tvd[:,col-1]

    # these are in days of the year
    day0= np.floor(tnew[0]) # first day
    dayN = np.ceil(np.max(tnew)) # last day

    # these are the times relative to time zero
    middle = tnew-day0

    # use the first day
    ii = tnew < (day0+1)
    leftTime = -(tnew[ii]-day0)
    leftY = ynew[ii]

    # now use the last day
    ii = tnew > (dayN-1)
    rightY = np.flip(ynew[ii])
    rightTime = tnew[ii] -day0 + 1 

    tmp= np.hstack((leftTime,middle)) ; 
    th = np.hstack((tmp,rightTime))

    tmp = np.hstack((leftY, ynew )) ; 
    h = np.hstack((tmp, rightY))

    # and sort it ...
    ii = np.argsort(th)
    th = th[ii] ; h = h[ii]

    th = th + day0 # add day0 back in

    # now fill the gaps ... 
    fillgap = 1/24 # one hour fake values
    # ???
    gap = 5/24 # up to five hour gap allowed before warning

    tnew =[] ; ynew =[]; faket = [];
    # fill in gaps using variables called tnew and ynew
    Ngaps = 0
    for i in range(1,len(th)):
        d= th[i]-th[i-1] # delta in time in units of days ?
        if (d > gap):
            x0 = th[i-1:i+1] ; h0 = h[i-1:i+1]

            # only print out the gap information the first time thru
            if col == 3:
                print('Gap on doy:', int(np.floor(x0[0])), ' lasting ', round(d*24,2), ' hours ')
            #print(d,x0,h0)
            Ngaps = Ngaps + 1
            f = scipy.interpolate.interp1d(x0,h0)
            # so this is fake data
            ttnew = np.arange(th[i-1]+fillgap, th[i], fillgap)
            yynew = f(ttnew)
            faket = np.append(faket, ttnew)
            # now append it to your real data
            tnew = np.append(tnew,ttnew)
            ynew = np.append(ynew,yynew)
        else:
            tnew = np.append(tnew,th[i])
            ynew = np.append(ynew,h[i])


    if (Ngaps > 3) and (col == 3):
        print('\nThis is a beta version of the rhdot/spline fit code - and does not')
        print('work well with gaps. You have been warned!\n')

    # sort again
    ii = np.argsort( tnew) 
    tnew = tnew[ii]
    ynew = ynew[ii]

    return tnew, ynew


def rhdot_correction2(station,fname,fname_new,pltit,outlierV,outlierV2,**kwargs):
    """
    Part two of subdaily.  It computes rhdot correction and interfrequency bias correction for RH 
    time series. This code assumes you have at least removed crude outliers in the previous section of the 
    subdaily code.

    Parameters
    ----------
    station : str
        4 char station name
    fname : list of str
        input filename(s) 
    fname_new : str
        output filename for results
    pltit : bool
        whether you want plots to the screen
    outlierV : float
       outlier criterion, in meters  used in first go thru
       if None, then use 3 sigma (which is the default)
    outlierV2 : float
       outlier criterion, in meters  used in second go thru
       if None, then use 3 sigma (which is the default)
    delta_out : float, optional
        seconds for smooth output
    txtdir : str
        if wanting to set your own output directory
    apply_if_corr : bool, optional
        whether you want to apply the IF correction
        default is true
    apply_rhdot : bool, optional
        whether you want to apply the rhdot correction
        default is true
    gap_min_val : float, optional
        gap allowed in last spline, in hours
    knots2 : int, optional
        a secondary knot value if you want the final output 
        to use a different one than the one used for outliers and 
        RH dot
    gap_flag : bool, option
        whether gaps are written as 999 in final output


    """
    # output will go to REFL_CODE/Files unless txtdir provided
    xdir = os.environ['REFL_CODE']

    # change units to day of year
    gap_min_val = kwargs.get('gap_min_val',6.0)
    gap_min_val = gap_min_val/24 # change to DOY units

    val = kwargs.get('txtdir',[])

    gap_flag = kwargs.get('gap_flag',False)
    print('gap flag ', gap_flag)

    if len(val) == 0:
        txtdir = xdir + '/Files/'
    else:
        txtdir = val

    year = kwargs.get('year',0)

    # extension now input
    extension = kwargs.get('extension','')

    delta_out = kwargs.get('delta_out',0)
    if (delta_out  == 0):
        splineout = False
    else:
        splineout = True

    if_corr = kwargs.get('if_corr',True)
    if if_corr:
        apply_if_corr = True
    else:
        apply_if_corr = False

    writecsv = kwargs.get('csv',False)


    apply_rhdot  = kwargs.get('apply_rhdot',True)
    if apply_rhdot:
        apply_rhdot_corr = True
    else:
        apply_rhdot_corr = False

    # probably a better way - but this is it for now ....
    gotit = kwargs.get('hires_figs',True)
    if gotit:
        hires_figs = True
        print('Requested high resolution figures')
    else:
        hires_figs = False

    fs = kwargs.get('fs',12)


#   how often do you want velocity computed (per day)
    perday = 24*20 # so every 3 minutes
    # making a knot every three hours ...
    # knots_per_day = 8
    knots_default = 8
    knots_per_day= kwargs.get('knots',8)


    knots2 = kwargs.get('knots2',knots_per_day)
    if knots2 is None:
        knots2_per_day = knots_per_day
    else:
        knots2_per_day = knots2

    print('knots2 ', knots2_per_day) 

    knots_test = kwargs.get('knots_test',0)
    if (knots_test== 0):
        knots_test = knots_per_day

    print('\n>>>>>>>>>>>>>>>>>>>> Entering second section of subdaily code <<<<<<<<<<<<<<<<<<<<<<<<')
    print('\nComputes rhdot correction and interfrequency bias correction for subdaily')

    # read in the lomb scargle values which are the output of gnssir i.e. the reflector heights
    nfiles = len(fname)
    if nfiles == 1:
        multiyear = False
        print('\nInput filename:', fname[0])
        tvd = np.loadtxt(fname[0],comments='%')
    else:
        multiyear = True
        for i in range(0,nfiles):
            print('\nInput filename:', fname[i])
            newl = np.loadtxt(fname[i], comments='%')
            if i == 0:
                tvd = newl
            else:
                tvd = np.vstack((tvd, newl))

    if len(tvd) == 0:
        print('empty input file')
        return

    print('\nOutput filename: ', fname_new)
    print('\nMinimum gap allowed in spline output, in day units', gap_min_val)

    # sort the data in time
    # 2024-jan-3 change to use MJD
    #ii = np.argsort( (tvd[:,1]+tvd[:,4]/24) ).T
    ii = np.argsort( tvd[:,15] ).T
    tvd = tvd[ii,:]


    # originally I used day of year.  Then i used MJD for multi-year and 
    # doy of year for dates within a year.... now i am trying to consolidate
    # entirely as MJD
    # use MJD
    th= tvd[:,15] ; h = tvd[:,2]
    tnew, ynew = sd.flipit3(tvd,3)

    #else:
    # use day of year for the spline fits
    #th= tvd[:,1] + tvd[:,4]/24; 
    #tnew, ynew = flipit(tvd,3)
    #    h = tvd[:,2]
    # try this instead of of fractional doy
    #    th = tvd[:,15]
    #    tnew, ynew = flipit2(tvd,3)


    # this is the edot factor - 
    # this was computed by gnssir as the mean of the tangent(eangles) over an arc,
    # divided by edot (the time rate of change of elevation angle, initially rad/sec,
    # but then converted to rad/hour).  Thus this parameter has units  rad/(rad/hour) >>>> hours
    # it is multiplied by RHDot in meters/hour, which gives you a correction value in meters
    xfac = tvd[:,12]


    Ndays = tnew.max()-tnew.min()
    numKnots = int(knots_per_day*(Ndays))
    print('First and last time values in the spline', '{0:8.3f} {1:8.3f} '.format (tnew.min(), tnew.max()) )
    print('Number of RH obs and Days ', len(h), np.round(Ndays,3))
    print('Average num of RH obs per day', '{0:5.1f} '.format (len(h)/Ndays) )
    print('Knots per day: ', knots_per_day, ' Number of knots: ', numKnots)
    # currently using 3 sigma
    print('Outlier criterion provided by user for the first splinefit (m):', outlierV)
    print('Outlier criterion provided by user for the second splinefit (m):', outlierV2)


    firstKnot_in_minutes = 15
    t1 = tnew.min()+firstKnot_in_minutes/60/24
    t2 = tnew.max()-firstKnot_in_minutes/60/24
    knots =np.linspace(t1,t2,num=numKnots)

    t, c, k = interpolate.splrep(tnew, ynew, s=0, k=3,t=knots,task=-1)


    spline = interpolate.BSpline(t, c, k, extrapolate=False)
    # this is to get  RHdot, evenly spaced data - units of days
    N = int(Ndays*perday)
    xx = np.linspace(tnew.min(), tnew.max(), N)
    spl_x = xx; spl_y = spline(xx)
    # these are spline values at the RH observation times
    spl_at_GPS_times = spline(th) 

    sd.mirror_plot(tnew,ynew,spl_x,spl_y,txtdir,station,th[0],th[-1])

    plot_begin = np.floor(np.min(th))
    plot_end =np.ceil(np.max(th)) 

    obsPerHour= perday/24

    tvel = spl_x[1:N]
    yvel = obsPerHour*np.diff(spl_y)

    rhdot_at_th = np.interp(th, tvel, yvel)
    # RH dot correction
    correction = xfac*rhdot_at_th

    # 2023 august 25
    if not apply_rhdot :
        print('You requested RHdot correction NOT be applied.')
        correctedRH = h
    else:
    # corrected reflector height
        correctedRH = h-correction

    # this is RH with the RHdot correction
    residual_before = h - spl_at_GPS_times
    residual_after = correctedRH - spl_at_GPS_times
    sigmaBefore = np.std(residual_before)
    sigmaAfter  = np.std(residual_after)

    label1 = 'w/o RHdot ' + str(np.round(sigmaBefore,3)) + 'm'
    label2 = 'w RHdot ' + str(np.round(sigmaAfter,3)) + 'm'

    # start the figure, convert from doy to MJD to obstime
    # not sure this is still used?
    mjd0 = g.fdoy2mjd(year,th[0])
    th_obs = sd.mjd_to_obstimes(mjd0 + th-th[0])

    #get datetime values
    th_obs = sd.mjd_to_obstimes(th)

    fig=plt.figure(figsize=(10,6))
    plt.subplot(2,1,1)
    plt.plot(th_obs, h, 'b.', label=label1,markersize=4)

    iw = (spl_x > th[0]) & (spl_x < th[-1])

    # change to datetime
    #if multiyear:
    spl_x_obs = sd.mjd_to_obstimes(spl_x[iw])
    plt.plot(spl_x_obs, spl_y[iw], 'c--', label='spline') 
    #else:
    #    tm_mjd = g.fdoy2mjd(year, spl_x[iw][0])
    #    spl_x_obs = sd.mjd_to_obstimes(tm_mjd + spl_x[iw] - spl_x[iw][0] )
    #    plt.plot(spl_x_obs, spl_y[iw], 'c--', label='spline') # otherwise wonky spline makes a goofy plot

    plt.plot(th_obs,correctedRH,'m.',label=label2,markersize=4)

    print('\nRMS no RHdot correction (m)', '{0:6.3f}'.format ( sigmaBefore))
    print('RMS w/ RHdot correction (m)', '{0:6.3f} \n'.format ( sigmaAfter ))

    plt.gca().invert_yaxis()
    plt.legend(loc="upper left")
    plt.ylabel('meters',fontsize=fs)
    plt.title( station.upper() + ' Reflector Heights',fontsize=fs)
    outlierstring = str(outlierV) + '(m) outliers'
    plt.yticks(fontsize=fs); plt.xticks(fontsize=fs)
    plt.grid()
    fig.autofmt_xdate()


    plt.subplot(2,1,2)
    plt.plot(th_obs, residual_after,'r.',label='all pts')
    plt.title('RH residuals')

    if outlierV is None:
    # use 3 sigma
        OutlierLimit = 3*sigmaAfter
        print('Using three sigma outlier criteria')
    else:
        OutlierLimit = float(outlierV)
        print('User-defined splinefit outlier value (m): ', OutlierLimit)

    tvd_bad = tvd[np.abs(residual_after) >  OutlierLimit, :]
    #tvd_bad = tvd[np.abs(residual_after) >  3*sigmaAfter, :]


    #kk = np.abs(residual_after) > 3*sigmaAfter
    kk = np.abs(residual_after) > OutlierLimit
    tvd_confused = tvd[kk,:]

    sd.writeout_spline_outliers(tvd_confused,txtdir,residual_after[kk],'outliers.spline.txt')

    # keep values within 3 sigma 
    #ii = np.abs(residual_after) < 3*sigmaAfter
    ii = np.abs(residual_after) <= OutlierLimit
    tvd_new = tvd[ii,:]
    correction_new = correction[ii]
    correctedRH_new = correctedRH[ii]
    NV = len(correctedRH)

    plt.plot(th_obs[ii], residual_after[ii],'b.',label='kept pts')
    plt.grid()
    plt.ylabel('meters',fontsize=fs)
    plt.legend(loc="upper left",fontsize=fs)
    plt.yticks(fontsize=fs); plt.xticks(fontsize=fs)
    fig.autofmt_xdate()

    if hires_figs:
        g.save_plot(txtdir + '/' + station + '_rhdot2.eps')
    else:
        g.save_plot(txtdir + '/' + station + '_rhdot2.png')

    # update the residual vector as well
    residual_after = residual_after[ii]

    # make the RHdot plot as well. if multiyear, Skip for now
    if not multiyear:
        sd.rhdot_plots(th,correction,rhdot_at_th, tvel,yvel,fs,station,txtdir,hires_figs,year)

    #writecsv = False ; 
    extraline = ''
    # write out the new solutions with RHdot and without 3 sigma outliers
    # this should be changed to take into account apply_rhdot_corr
    write_subdaily(fname_new,station,tvd_new,writecsv,extraline,newRH=correctedRH_new, RHdot_corr=correction_new)
    nr,nc = tvd_new.shape
    print('Percent of observations removed:', round(100*(NV-nr)/NV,2))

    # now make yet another vector - this time to apply bias corrections
    # this is horrible code and needs to be fixed
    biasCorrected_RH = correctedRH_new

    print('----------------------------------------------------------\n')
    print('Check inter-frequency biases for GPS, Glonass, Galileo, and Beidou \n')

    if 1 not in tvd_new[:,10]:
        print('Biases are computed with respect to the average of all constellations')
        L1exist = False
    else:
        print('Biases will be computed with respect to L1 GPS')
        L1exist = True

    tvd_new = np.loadtxt(fname_new,comments='%')
    nr,nc = tvd_new.shape
    onecol = np.zeros((nr,1)) # 

    # add a column for the IF correction
    tvd_new = np.hstack((tvd_new,onecol))
    nr,nc = tvd_new.shape

    print('Freq  Bias  Sigma   NumObs ')
    print('       (m)   (m)       ')
    for f in [1, 2, 20, 5, 101, 102, 201, 205,206,207,208,302,306,307]:
        ff = (tvd_new[:,10] == f)
        ret = residual_after[ff]
        if len(ret) > 0:
            bias = float(np.mean(ret))
            if f == 1: # save this bias
                L1bias = bias
            if L1exist: # apply L1bias so that everything is relative to L1
                bias = bias - L1bias

            tvd_new[ff,24] = tvd_new[ff,22] - bias

            # print stats to the screen
            sig = float(np.std(ret))
            print('{0:3.0f} {1:7.3f} {2:7.3f} {3:6.0f}'.format (f, bias, sig, len(ret) ) )


    if not apply_if_corr:
        print('You chose not to correct for IF biases.')
        print('Since the spline fit is written out after this stage of the ')
        print('code, please submit a pull request to changes the logic. Exiting')
        #if pltit:
        #    plt.show()
        return tvd_new, correction
    print('----------------------------------------------------------\n')

    # now try to write the bias corrected values
    # first attempt was wrong because i forgot to sort the corrected column in biasCorrected_RH
    #bias_corrected_filename = fname_new + 'IF'; extraline = ''; writecsv = False
    biasCor_rh = tvd_new[:,24]
    #write_subdaily(bias_corrected_filename,station,tvd_new, writecsv,extraline, newRH_IF=biasCor_rh)

    # now use the bias corrected value

 
    # add mirror time series at beginning and end of series for spline
    th = tvd_new[:,15] # use MJD now
    column = 25 # tells the code which column to use. 
    tnew, ynew = sd.flipit3(tvd_new,column)
    #else:
    #    th = tvd_new[:,1] + tvd_new[:,4]/24 # days of year, fractional
    #    column = 25 # tells the code which column to use. 
    #    tnew, ynew = flipit(tvd_new,column)

    Ndays = tnew.max()-tnew.min()
    #print('trying knots2')
    knots_per_day = knots2_per_day

    numKnots = int(knots_per_day*(Ndays))
    #

    firstKnot_in_minutes = 15
    t1 = tnew.min()+firstKnot_in_minutes/60/24
    t2 = tnew.max()-firstKnot_in_minutes/60/24
    knots =np.linspace(t1,t2,num=numKnots)


    t, c, k = interpolate.splrep(tnew, ynew, s=0, k=3,t=knots,task=-1)
    # compute spline - use for times th
    spline = interpolate.BSpline(t, c, k, extrapolate=False)
    #print('first and last th value - but for what exactly?')


    # calculate spline values at GPS time tags
    spline_at_GPS = spline(th)
    # for the plot I did 30 minutes - but that is not the same as what
    # is used for the final output.  Might as well have them be the same
    half_hourly = int(48*(th[-1] - th[0]))
    # evenly spaced spline values for the plot
    # I do not think these variables are used anymore

    #print(th[0], th[-1], half_hourly)
    th_even = np.linspace(th[0], th[-1],half_hourly ); 
    spline_whole_time = spline(th_even)

    newsigma = np.std(biasCor_rh-spline_at_GPS)
    strsig = str(round(newsigma,3)) + '(m)'

    if outlierV2 is None:
        # use 3 sigma to find outliers
        OutlierLimit = 3*sigmaAfter
        print('Using three sigma outlier criteria')
        ii = np.abs(biasCor_rh -spline_at_GPS)/newsigma > 3
        jj = np.abs(biasCor_rh -spline_at_GPS)/newsigma <= 3
    else:
        OutlierLimit = float(outlierV2)
        print('User-defined splinefit outlier value (m): ', OutlierLimit)
        # throw out
        ii = np.abs(biasCor_rh -spline_at_GPS) > OutlierLimit
        # points to keep
        jj = np.abs(biasCor_rh -spline_at_GPS) <= OutlierLimit

    
    # make the plot externally now
    badpoints2 = sd.subdaily_resids_last_stage(station, year, th, biasCor_rh, spline_at_GPS, 
                                               fs, strsig, hires_figs,txtdir, ii,jj,th_even, spline_whole_time)

    # pick up the orthometric height from the gnssir_analysis json
    H0 = sd.find_ortho_height(station,extension)
    # this writes out spline file and makes plot .... 
    sd.RH_ortho_plot2( station, H0, year, txtdir, fs, th[jj],biasCor_rh[jj],gap_min_val,th,spline,delta_out,writecsv,gap_flag)
    print('\nRMS with frequency biases and RHdot taken out (m) ', np.round(newsigma,3) , '\n' )


    # write out the files with RH dot and IF bias corrected - and again without 3 sigma outliers
    bias_corrected_filename = fname_new + 'IF'; extraline = ''; 
    biasCor_rh = tvd_new[jj,24]
    write_subdaily(bias_corrected_filename,station,tvd_new[jj,:], writecsv,extraline, newRH_IF=biasCor_rh)

    new_outliers = tvd_new[ii,:]

    # write outliers to a file ... again ... 
    sd.writeout_spline_outliers(new_outliers,txtdir,badpoints2,'outliers.spline2.txt')


    return tvd, correction


def my_percentile(rh,p1, p2):
    """
    numpy percentile was crashing docker build
    this is a quick work around

    Parameters
    ----------
    rh : numpy array
        reflector heights, but could be anything really
    p1 : float
        low percentage (from 0-1)
    p2 : float
        high percentage (from 0-1)

    Returns
    -------
    low : float
        low value (using input percentile)
    highv : float
        high value (using input percentile)

    """

    sorted_rh = np.sort(rh)
    N = len(sorted_rh)
    N1 = int(np.round(p1*N))
    N2 = int(np.round(p2*N))

    # calculate the low and high values at these percentiles
    lowv = sorted_rh[N1] 
    highv = sorted_rh[N2]

    return  lowv, highv

def flipit2(tvd,col):
    """
    take RH values from the first and last day and attaches
    them as fake data to make the spline fit stable.  
    Also fill the temporal gaps with fake data

    This version uses MJD rather than day of year for x-axis

    Parameters
    ----------
    tvd : numpy array of floats
        output of LSP runs. 
    col : integer
        column number (in normal speak) of the RH results
        in python-speak, this has one subtracted

    Returns
    -------
    tnew : numpy array of floats
        time in days of year
    ynew : numpy array
        RH in meters 

    """
    nr,nc = np.shape(tvd)
    #print(nr,nc)
    # sort it just to make sure ...
    #tnew = tvd[:,1] + tvd[:,4]/24
    # use MJD
    tnew = tvd[:,15]
    # change from normal columns to python columns
    ynew = tvd[:,col-1]

    # these are in days of the year
    day0= np.floor(tnew[0]) # first day
    dayN = np.ceil(np.max(tnew)) # last day

    # these are the times relative to time zero
    middle = tnew-day0

    # use the first day
    ii = tnew < (day0+1)
    leftTime = -(tnew[ii]-day0)
    leftY = ynew[ii]

    # now use the last day
    ii = tnew > (dayN-1)
    rightY = np.flip(ynew[ii])
    rightTime = tnew[ii] -day0 + 1 

    tmp= np.hstack((leftTime,middle)) ; 
    th = np.hstack((tmp,rightTime))

    tmp = np.hstack((leftY, ynew )) ; 
    h = np.hstack((tmp, rightY))

    # and sort it ...
    ii = np.argsort(th)
    th = th[ii] ; h = h[ii]

    th = th + day0 # add day0 back in

    # now fill the gaps ... 
    fillgap = 1/24 # one hour fake values
    # ???
    gap = 5/24 # up to five hour gap allowed before warning

    tnew =[] ; ynew =[]; faket = [];
    # fill in gaps using variables called tnew and ynew
    Ngaps = 0
    for i in range(1,len(th)):
        d= th[i]-th[i-1] # delta in time in units of days ?
        if (d > gap):
            x0 = th[i-1:i+1] ; h0 = h[i-1:i+1]

            # only print out the gap information the first time thru
            if col == 3:
                print('Gap on MJD:', int(np.floor(x0[0])), ' lasting ', round(d*24,2), ' hours ')
            #print(d,x0,h0)
            Ngaps = Ngaps + 1
            f = scipy.interpolate.interp1d(x0,h0)
            # so this is fake data
            ttnew = np.arange(th[i-1]+fillgap, th[i], fillgap)
            yynew = f(ttnew)
            faket = np.append(faket, ttnew)
            # now append it to your real data
            tnew = np.append(tnew,ttnew)
            ynew = np.append(ynew,yynew)
        else:
            tnew = np.append(tnew,th[i])
            ynew = np.append(ynew,h[i])


    if (Ngaps > 3) and (col == 3):
        print('\nThis is a beta version of the rhdot/spline fit code - and does not')
        print('work well with gaps. You have been warned!\n')

    # sort again
    ii = np.argsort( tnew) 
    tnew = tnew[ii]
    ynew = ynew[ii]

    return tnew, ynew




import numpy as np
import gnssrefl.gps as g
import os
import subprocess
import sys

import gnssrefl.xnmeasnr as xnmeasnr


def decipher_argt(station,filename,idec,snrname,orbfile,recx,csnr,year,month,day):
    """
    This is an attempt to properly model the satellite orbits.  It uses GNSS orbits
    from the GFZ and Fortran to compute azimuth and elevation angle.  
    Right now it is L1 only, but does allow Galileo, GPS, and Glonass.
    It does require that someone send a proper station location.

    This code was written specifically for a dataset collected in Argentina.
    It is not in NMEA format but was parsed from it.


    Parameters
    ----------
    station : str
        4 char id
    filename : str
        NMEA output from Argentina
    idec : int
        decimation interval, sec
    snrname : str
        ultimate output file
    orbfile : str
        sp3 filename
    recx : list of floats
        Cartesian station coordinates in meters
    csnr: str
        2 ch snr file choice, i.e. '66' or '99'
    year : int
        full yaer
    month : int
        calendar month
    day : int
        calendar day

    """
    if (len(snrname) > 132) or (len(orbfile) > 132):
        print('The orbit or SNR file name is too long.')
        print('Make your environment variable names shorter.')
        sys.exit()
    t = [] ; prn = [] ; az = [] ; elv = []; snr = [] ; freq = []
    # i do not know why the people that this code are using strings 
    # for frequency ... but so as to not break the code
    igal = 0; iglo = 0; igps = 0

    with open(filename, 'r') as f:
        lines = f.read().splitlines(True)
        for i in  range(0,86400):
            try:
                line = lines[i]
                obs = line.split(';')
                N = len(obs)
                time = obs[0]
                Hr = int(time[11:13] )
                minutes = int(time[14:16] )
                seconds = int(time[17:19] )
                sod =  Hr*3600 + minutes*60 + seconds

                for el in range(4,N-4,4):
                    sat = int(obs[el])
                    eleA = float(obs[el+3])
                    azimA = float(obs[el+2])
                    dat = float(obs[el+1])
                    if (eleA < 35) & (dat > 0):
                        if (sat < 33):
                            igps = igps + 1
                            prn.append(sat)
                            freq.append(1)
                            t.append(sod) ; snr.append(dat) ; az.append(azimA) ; elv.append(eleA)
                        if (sat > 64) & (sat <=88):
                            iglo = iglo + 1
                            newsat = sat + 100  -64 
                            prn.append(newsat)
                            t.append(sod) ; snr.append(dat) ; az.append(azimA) ; elv.append(eleA)
                            freq.append(1)
                        if (sat > 200) & (sat <=240):
                            igal = igal + 1
                            prn.append(sat)
                            freq.append(1)
                            t.append(sod) ; snr.append(dat) ; az.append(azimA) ; elv.append(eleA)

            except:
                iii = 0
    #print(igal,igps,iglo)
    message = 'None '
    xdir = os.environ['REFL_CODE']
    logdir = xdir + '/logs'
    # make sure this directory exists
    if not os.path.isdir(logdir):
        subprocess.call(['mkdir', logdir])

    # temporary file - will be deleted
    outputfile = logdir + '/' + station + 'tmp.txt'
    # write to file tod, iprn, s1, freq
    fout = open(outputfile, 'w+')
    fout.write('{0:15.4f}{1:15.4f}{2:15.4f} \n'.format(recx[0], recx[1],recx[2]) )
    fout.write('{0:6.0f}{1:6.0f}{2:6.0f} \n'.format(year, month, day) )
    for i in range(0,len(t)):
        # might as well decimate
        if ( (int(t[i]) % idec) == 0):
            fout.write('{0:8.0f} {1:3.0f} {2:6.2f} {3:1.0f} \n'.format(t[i], prn[i], snr[i],freq[i]) )
    fout.close()

    errorlog = logdir + '/' + station + '_nmea2snr_error.txt'

    in1 = g.binary(outputfile)
    in2 = g.binary(snrname)
    in3 = g.binary(orbfile)
    in4 = g.binary(csnr)
    in5 = g.binary(errorlog)
    xnmeasnr.foo(in1,in2,in3,in4,in5)
    subprocess.call(['rm',outputfile])


def new_azel(station,tmpfile,snrname,orbfile,csnr):
    """
    This is an attempt to properly model the satellite orbits.  It uses GNSS orbits
    from the GFZ and Fortran to compute azimuth and elevation angle.  
    Right now it is L1 only, but does allow Galileo, GPS, and Glonass.
    It does require that someone send a proper station location.


    Parameters
    ----------
    tmpfile: str
        NMEA output from Argentina
    snrname : str
        ultimaet output file
    orbfile : str
        sp3 filename
    csnr: str
        2 ch snr file choice, i.e. '66' or '99'

    """
    if (len(snrname) > 132) or (len(orbfile) > 132):
        print('The orbit or SNR file name is too long.')
        print('Make your environment variable names shorter.')
        sys.exit()
    # make sure this directory exists
    if not os.path.isdir('logs'):
        subprocess.call(['mkdir', 'logs'])

    errorlog = 'logs/' + station + '_nmea2snr_error.txt'

    in1 = g.binary(tmpfile)
    in2 = g.binary(snrname)
    in3 = g.binary(orbfile)
    in4 = g.binary(csnr)
    in5 = g.binary(errorlog)
    xnmeasnr.foo(in1,in2,in3,in4,in5)
    print('removing temporary file')
    subprocess.call(['rm',tmpfile])



# -*- coding: utf-8 -*-
import argparse
import datetime
import os
import requests
import subprocess
import sys
import gnssrefl.gps as g

from gnssrefl.utils import validate_input_datatypes, str2bool


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("station", help="station name, e.g. 5970026", type=str)
    parser.add_argument("-plt", default=None, help="Optional plot to screen", type=str)
    parser.add_argument("-output", default=None, help="Optional filename for wsv", type=str)
    args = parser.parse_args().__dict__


    # convert all expected boolean inputs from strings to booleans
    boolean_args = ['plt']
    args = str2bool(args, boolean_args)


    # only return a dictionary of arguments that were added from the user - all other defaults will be set in code below
    return {key: value for key, value in args.items() if value is not None}


def download_wsv(station: str, plt: bool = True, output: str = None):
    """
    Downloads and saves WSV (Germany) tide gauge files

    Parameters
    ----------
    station : str
        station name 
    plt: bool, optional
        plot comes to the screen
        default is None
    output: str, optional
        output filename which is stored in $REFL_CODE/Files
        if not set, it uses station.txt

    """
    g.check_environ_variables()
    # set up the address for the API call
    xdir = os.environ['REFL_CODE']
    outputdir = xdir + '/Files'  
    if not os.path.isdir(outputdir):
        print('making ', outputdir)
        subprocess.call(['mkdir',outputdir])

    newurl = 'https://www.pegelonline.wsv.de/webservices/rest-api/v2/stations/' + station + '/W/measurements.json?start=P30D'

    try:
        data = requests.get(newurl).json()
    except:
        print('Major issue with download. Is the internet off?')
        sys.exit()

    N= len(data)
    if (N <= 2) :
        print(data)
        sys.exit()
    thetime = []; sealevel = [] ; obstimes = [] ; pt = 0
    if output is None:
        # use default
        outfile = outputdir + '/' + station + '.txt'
    else:
        outfile = outputdir + '/' + output

    # open the file
    print('File written to :', outfile)
    fout = open(outfile,'w+')
    fout.write("{0:s} \n".format('%' + ' WSV Station: ' + station ))
    fout.write("%YYYY MM DD  HH MM SS  Water(m) DOY  MJD     \n")
    fout.write("% 1   2  3   4  5  6     7      8     9  \n")


    for i in range(0, N):
        sl = float(data[i]['value'])/100 # change to meters
        t = data[i]['timestamp']
        o=datetime.datetime.fromisoformat(t)
        #print(o)
        ts = datetime.datetime.utctimetuple(o)
        #    print(ts.tm_year, ts.tm_mon, ts.tm_mday, ts.tm_hour, ts.tm_min, ts.tm_sec, ts.tm_yday)
        year = ts.tm_year ; mm  = ts.tm_mon ; dd =  ts.tm_mday
        hh = ts.tm_hour ; minutes = ts.tm_min ; sec = ts.tm_sec
        doy = ts.tm_yday
        # calcualte MJD
        m, f = g.mjd(year, mm, dd, hh, minutes, sec)
        mjd = m + f;
        thetime.append(mjd); sealevel.append(sl)
            #print(year,mm,dd,hh,minutes,sec)
        bigT = datetime.datetime(year=year, month=mm, day=dd, hour=hh, minute=minutes, second=sec)
        obstimes.append(bigT)

        fout.write(" {0:4.0f} {1:2.0f} {2:2.0f} {3:2.0f} {4:2.0f} {5:2.0f} {6:7.3f} {7:3.0f} {8:16.7f} \n".format(year, 
                mm, dd, hh, minutes, sec, sl, doy, mjd))
    # close the file
    fout.close()

    if plt:
        g.quickp(station,obstimes,sealevel)

def main():
    args = parse_arguments()
    download_wsv(**args)


if __name__ == "__main__":
    main()
